from unittest import mock

from pyramid import httpexceptions

from kinto.core.permission.memory import Permission
from kinto.core.resource import Resource

from . import BaseTest


class PermissionTest(BaseTest):
    resource_class = Resource

    def setUp(self):
        self.permission = Permission()
        super().setUp()

    def get_request(self):
        request = super().get_request()
        request.registry.permission = self.permission
        return request


class PluralEndpointPermissionTest(PermissionTest):
    def setUp(self):
        super().setUp()
        self.result = self.resource.plural_get()

    def test_permissions_are_not_provided_in_plural_get(self):
        self.assertNotIn("permissions", self.result)

    def test_permissions_are_not_provided_in_plural_delete(self):
        result = self.resource.plural_delete()
        self.assertNotIn("permissions", result)


class ObtainObjectPermissionTest(PermissionTest):
    def setUp(self):
        super().setUp()
        obj = self.resource.model.create_object({})
        object_id = obj["id"]
        object_uri = "/articles/{}".format(object_id)
        self.permission.add_principal_to_ace(object_uri, "read", "basicauth:bob")
        self.permission.add_principal_to_ace(object_uri, "read", "account:readonly")
        self.permission.add_principal_to_ace(object_uri, "write", "basicauth:bob")
        self.resource.object_id = object_id
        self.resource.request.validated["body"] = {"data": {}}
        self.resource.request.path = object_uri

    def test_permissions_are_provided_in_object_get(self):
        result = self.resource.get()
        self.assertIn("permissions", result)

    def test_permissions_are_provided_in_object_put(self):
        result = self.resource.put()
        self.assertIn("permissions", result)

    def test_permissions_are_provided_in_object_patch(self):
        result = self.resource.patch()
        self.assertIn("permissions", result)

    def test_permissions_are_not_provided_in_object_delete(self):
        result = self.resource.delete()
        self.assertNotIn("permissions", result)

    def test_permissions_gives_lists_of_principals_per_ace(self):
        result = self.resource.get()
        permissions = result["permissions"]
        self.assertEqual(sorted(permissions["read"]), ["account:readonly", "basicauth:bob"])
        self.assertEqual(sorted(permissions["write"]), ["basicauth:bob"])

    def test_permissions_are_hidden_if_user_has_only_read_permission(self):
        self.resource.model.current_principal = "account:readonly"
        self.resource.model.prefixed_principals = []
        result = self.resource.get()
        self.assertEqual(result["permissions"], {})


class SpecifyObjectPermissionTest(PermissionTest):
    def setUp(self):
        super().setUp()
        self.obj = self.resource.model.create_object({})
        object_id = self.obj["id"]
        self.object_uri = "/articles/{}".format(object_id)
        self.permission.add_principal_to_ace(self.object_uri, "read", "account:readonly")
        self.resource.request.matchdict["id"] = object_id
        self.resource.request.validated["body"] = {"data": {}}
        self.resource.request.path = self.object_uri

    def test_write_permission_is_given_to_creator_on_post(self):
        self.resource.context.object_uri = "/articles"
        self.resource.request.method = "POST"
        result = self.resource.plural_post()
        self.assertEqual(sorted(result["permissions"]["write"]), ["basicauth:bob"])

    def test_write_permission_is_given_to_put(self):
        self.resource.request.method = "PUT"
        result = self.resource.put()
        permissions = result["permissions"]
        self.assertEqual(sorted(permissions["write"]), ["basicauth:bob"])

    def test_write_permission_is_given_to_anonymous(self):
        request = self.get_request()
        # Simulate an anonymous PUT
        request.method = "PUT"
        request.validated = {**self.resource.request.validated, "body": {"data": {**self.obj}}}
        request.prefixed_userid = None
        request.matchdict = {"id": self.obj["id"]}
        resource = self.resource_class(request=request, context=self.get_context())
        result = resource.put()
        self.assertIn("system.Everyone", result["permissions"]["write"])

    def test_permissions_can_be_specified_in_plural_post(self):
        perms = {"write": ["jean-louis"]}
        self.resource.request.method = "POST"
        self.resource.context.object_uri = "/articles"
        self.resource.request.validated["body"] = {"data": {}, "permissions": perms}
        result = self.resource.plural_post()
        self.assertEqual(sorted(result["permissions"]["write"]), ["basicauth:bob", "jean-louis"])

    def test_permissions_are_replaced_with_put(self):
        perms = {"write": ["jean-louis"]}
        self.resource.request.validated["body"]["permissions"] = perms
        self.resource.request.method = "PUT"
        result = self.resource.put()
        # In setUp() 'read' was set on this object.
        # PUT had got rid of it:
        self.assertNotIn("read", result["permissions"])

    def test_permissions_are_modified_with_patch(self):
        perms = {"write": ["jean-louis"]}
        self.resource.request.validated["body"] = {"permissions": perms}
        self.resource.request.method = "PATCH"
        result = self.resource.patch()
        permissions = result["permissions"]
        self.assertEqual(sorted(permissions["read"]), ["account:readonly"])
        self.assertEqual(sorted(permissions["write"]), ["basicauth:bob", "jean-louis"])

    def test_permissions_can_be_removed_with_patch_but_keep_current_user(self):
        self.permission.add_principal_to_ace(self.object_uri, "write", "jean-louis")

        perms = {"write": []}
        self.resource.request.validated["body"] = {"permissions": perms}
        self.resource.request.method = "PATCH"
        result = self.resource.patch()
        permissions = result["permissions"]
        (self.assertEqual(sorted(permissions["read"]), ["account:readonly"]),)
        self.assertEqual(sorted(permissions["write"]), ["basicauth:bob"])

    def test_permissions_can_be_removed_with_patch(self):
        self.permission.add_principal_to_ace(self.object_uri, "write", "jean-louis")

        perms = {"read": []}
        self.resource.request.validated["body"] = {"permissions": perms}
        self.resource.request.method = "PATCH"
        result = self.resource.patch()
        self.assertNotIn("read", result["permissions"])
        self.assertEqual(sorted(result["permissions"]["write"]), ["basicauth:bob", "jean-louis"])

    def test_412_errors_do_not_put_permission_in_object(self):
        self.resource.request.validated["header"] = {"If-Match": 1234567}  # invalid
        try:
            self.resource.put()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertEqual(
            error.json["details"]["existing"],
            {"id": self.obj["id"], "last_modified": self.obj["last_modified"]},
        )


class DeletedObjectPermissionTest(PermissionTest):
    def setUp(self):
        super().setUp()
        obj = self.resource.model.create_object({})
        self.resource.object_id = object_id = obj["id"]
        self.object_uri = "/articles/{}".format(object_id)
        self.resource.request.route_path.return_value = self.object_uri
        self.resource.request.path = self.object_uri
        self.permission.add_principal_to_ace(self.object_uri, "read", "fxa:user")

    def test_permissions_are_deleted_when_object_is_deleted(self):
        self.resource.delete()
        principals = self.permission.get_object_permission_principals(self.object_uri, "read")
        self.assertEqual(len(principals), 0)

    def test_permissions_are_deleted_when_plural_is_deleted(self):
        self.resource.context.on_plural_endpoint = True
        self.resource.plural_delete()
        principals = self.permission.get_object_permission_principals(self.object_uri, "read")
        self.assertEqual(len(principals), 0)


class GuestPluralEndpointTest(PermissionTest):
    def setUp(self):
        super().setUp()
        object1 = self.resource.model.create_object({"letter": "a"})
        object2 = self.resource.model.create_object({"letter": "b"})
        object3 = self.resource.model.create_object({"letter": "c"})

        uri1 = "/articles/{}".format(object1["id"])
        uri2 = "/articles/{}".format(object2["id"])
        uri3 = "/articles/{}".format(object3["id"])

        self.permission.add_principal_to_ace(uri1, "read", "fxa:user")
        self.permission.add_principal_to_ace(uri2, "read", "group")
        self.permission.add_principal_to_ace(uri3, "read", "jean-louis")

        self.resource.context.shared_ids = [object1["id"], object2["id"]]

    def test_plural_is_filtered_for_current_guest(self):
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 2)

    def test_guest_plural_get_can_be_filtered(self):
        self.resource.request.validated["querystring"] = {"letter": "a"}
        with mock.patch.object(self.resource, "is_known_field"):
            result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 1)

    def test_guest_plural_get_is_empty_if_no_object_is_shared(self):
        self.resource.context.shared_ids = ["tata lili"]
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_permission_backend_is_not_queried_if_not_guest(self):
        self.resource.context.shared_ids = None
        self.resource.request.registry.permission = None  # would fail!
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 3)

    def test_unauthorized_error_if_resource_does_not_exist(self):
        pass


class GuestPluralDeleteTest(PermissionTest):
    def setUp(self):
        super().setUp()
        object1 = self.resource.model.create_object({"letter": "a"})
        object2 = self.resource.model.create_object({"letter": "b"})
        object3 = self.resource.model.create_object({"letter": "c"})
        object4 = self.resource.model.create_object({"letter": "d"})

        uri1 = "/articles/{}".format(object1["id"])
        uri2 = "/articles/{}".format(object2["id"])
        uri3 = "/articles/{}".format(object3["id"])
        uri4 = "/articles/{}".format(object4["id"])

        self.permission.add_principal_to_ace(uri1, "read", "fxa:user")
        self.permission.add_principal_to_ace(uri2, "write", "fxa:user")
        self.permission.add_principal_to_ace(uri3, "write", "group")
        self.permission.add_principal_to_ace(uri4, "write", "jean-louis")

        self.resource.context.shared_ids = [object2["id"], object3["id"]]
        self.resource.request.method = "DELETE"

    def get_request(self):
        request = super().get_request()
        # RouteFactory must be aware of DELETE to query 'write' permission.
        request.method = "DELETE"
        return request

    def test_plural_is_filtered_for_current_guest(self):
        self.resource.request.path = "/articles"
        result = self.resource.plural_delete()
        self.assertEqual(len(result["data"]), 2)

    def test_guest_plural_delete_can_be_filtered(self):
        self.resource.request.validated["querystring"] = {"letter": "b"}
        with mock.patch.object(self.resource, "is_known_field"):
            result = self.resource.plural_delete()
        self.assertEqual(len(result["data"]), 1)
        objects = self.resource.model.get_objects()
        self.assertEqual(len(objects), 3)

    def test_guest_cannot_delete_objects_if_not_allowed(self):
        self.resource.context.shared_ids = ["tata lili"]
        result = self.resource.plural_delete()
        self.assertEqual(len(result["data"]), 0)
        objects = self.resource.model.get_objects()
        self.assertEqual(len(objects), 4)



import re
import unittest

from playwright.sync_api import expect, sync_playwright


baseUrl = "http://localhost:8888/v1/"
auth = {"user": "user", "password": "p4ssw0rd"}

browser = sync_playwright().start().firefox.launch()
context = browser.new_context(base_url=baseUrl)
page = browser.new_page()


class BrowserTest(unittest.TestCase):
    def setUp(self):
        request = context.request

        request.post("accounts", data={"data": {"id": auth["user"], "password": auth["password"]}})

    def test_login_and_view_home_page(self):
        page.goto(f"{baseUrl}admin/")

        expect(page).to_have_title(re.compile("Kinto Administration"))

        page.get_by_label("Kinto Account Auth").click()
        txtUsername = page.get_by_label(re.compile("Username"))
        txtPassword = page.get_by_label(re.compile("Password"))

        txtUsername.fill(auth["user"])
        txtPassword.fill(auth["password"])
        page.get_by_text(re.compile("Sign in using Kinto Account Auth")).click()

        expect(page.get_by_text("Kinto Administration")).to_be_visible()
        expect(page.get_by_text("project_name")).to_be_visible()
        expect(page.get_by_text("project_version")).to_be_visible()
        expect(page.get_by_text("http_api_version")).to_be_visible()
        expect(page.get_by_text("project_docs")).to_be_visible()



import re

import colander
from pyramid.settings import aslist

from kinto.authorization import PERMISSIONS_INHERITANCE_TREE
from kinto.core import resource
from kinto.core import utils as core_utils
from kinto.core.storage import Sort
from kinto.core.storage.memory import extract_object_set


def allowed_from_settings(settings, principals):
    """Returns every permissions allowed from settings for the current user.
    :param settings dict: app settings
    :param principals list: list of principals of current user
    :rtype: dict

    Result example::

        {
            "bucket": {"write", "collection:create"},
            "collection": {"read"}
        }

    XXX: This helper will be useful for Kinto/kinto#894
    """
    # Select settings about explicit permissions set on resources
    # bucket_write_principals = ... --> {("bucket", "write"): ["account:admin"]}
    perms_settings = {
        tuple(k.split("_", 3)[:2]): aslist(v)
        for k, v in settings.items()
        if re.match("(.+)_(create|write|read)_principals", k)
    }

    from_settings = {}
    for (resource_name, permission), allowed_principals in perms_settings.items():
        # Keep the known permissions only.
        if resource_name not in PERMISSIONS_INHERITANCE_TREE.keys():
            continue
        # Keep the permissions of the current user only.
        if not bool(set(principals) & set(allowed_principals)):
            continue
        # ``collection_create_principals`` means ``collection:create`` in bucket.
        if permission == "create":
            permission = f"{resource_name}:{permission}"
            resource_name = {  # resource parents.
                "collection": "bucket",
                "group": "bucket",
                "record": "collection",
                "bucket": "root",
                "account": "root",
            }.get(resource_name, "")
        # Store them in a convenient way.
        from_settings.setdefault(resource_name, set()).add(permission)
    return from_settings


class PermissionsModel:
    id_field = "id"
    modified_field = "last_modified"
    deleted_field = "deleted"

    def __init__(self, request):
        self.request = request

    def timestamp(self, parent_id=None):
        return 0

    def get_objects(
        self,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        parent_id=None,
    ):
        objects, _ = self._get_objects(
            filters=filters,
            sorting=sorting,
            pagination_rules=pagination_rules,
            limit=limit,
            include_deleted=include_deleted,
            parent_id=parent_id,
        )
        return objects

    def count_objects(self, filters=None, parent_id=None):
        _, count = self._get_objects(filters=filters, parent_id=parent_id)
        return count

    def _get_objects(
        self,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        parent_id=None,
    ):
        # Invert the permissions inheritance tree.
        perms_descending_tree = {}
        for on_resource, tree in PERMISSIONS_INHERITANCE_TREE.items():
            for obtained_perm, obtained_from in tree.items():
                for from_resource, perms in obtained_from.items():
                    for perm in perms:
                        perms_descending_tree.setdefault(from_resource, {}).setdefault(
                            perm, {}
                        ).setdefault(on_resource, set()).add(obtained_perm)

        # Obtain current principals.
        principals = self.request.prefixed_principals

        # Query every possible permission of the current user from backend.
        backend = self.request.registry.permission
        perms_by_object_uri = backend.get_accessible_objects(principals)

        # Check settings for every allowed resources.
        from_settings = allowed_from_settings(self.request.registry.settings, principals)

        # Add additional resources and permissions defined in settings/plugins
        for root_perm in from_settings.get("root", []):
            resource_name, _ = root_perm.split(":")
            perms_by_object_uri.setdefault("/", set()).add(root_perm)
            perms_descending_tree.setdefault("root", {}).update({root_perm: {"root": {root_perm}}})

        # Expand permissions obtained from backend with the object URIs that
        # correspond to permissions allowed from settings.
        allowed_resources = {"bucket", "collection", "group"} & set(from_settings.keys())
        if allowed_resources:
            storage = self.request.registry.storage
            every_bucket = storage.list_all(parent_id="", resource_name="bucket")
            for bucket in every_bucket:
                bucket_uri = "/buckets/{id}".format_map(bucket)
                for res in allowed_resources:
                    resource_perms = from_settings[res]
                    # Bucket is always fetched.
                    if res == "bucket":
                        perms_by_object_uri.setdefault(bucket_uri, set()).update(resource_perms)
                        continue
                    # Fetch bucket collections and groups.
                    # XXX: wrong approach: query in a loop!
                    every_subobjects = storage.list_all(parent_id=bucket_uri, resource_name=res)
                    for subobject in every_subobjects:
                        subobj_uri = bucket_uri + f"/{res}s/{subobject['id']}"
                        perms_by_object_uri.setdefault(subobj_uri, set()).update(resource_perms)

        entries = []
        for object_uri, perms in perms_by_object_uri.items():
            try:
                # Obtain associated res from object URI
                resource_name, matchdict = core_utils.view_lookup(self.request, object_uri)
            except ValueError:
                # Skip permissions entries that are not linked to an object URI
                continue

            # For consistency with event payloads, if resource has an id,
            # prefix it with its resource name
            if "id" in matchdict:
                matchdict[resource_name + "_id"] = matchdict["id"]

            # The imaginary "root" resource gets mapped to the hello
            # view. Handle it explicitly.
            if resource_name == "hello":
                resource_name = "root"

            # Expand implicit permissions using descending tree.
            permissions = set(perms)
            for perm in perms:
                obtained = perms_descending_tree[resource_name][perm]
                # Related to same resource only and not every sub-objects.
                # (e.g "bucket:write" gives "bucket:read" but not "group:read")
                permissions |= obtained[resource_name]

            entry = dict(
                uri=object_uri,
                resource_name=resource_name,
                permissions=list(permissions),
                **matchdict,
            )
            entries.append(entry)

        return extract_object_set(
            entries,
            filters=filters,
            sorting=sorting,
            id_field="uri",
            pagination_rules=pagination_rules,
            limit=limit,
        )


class PermissionsSchema(resource.ResourceSchema):
    uri = colander.SchemaNode(colander.String())
    resource_name = colander.SchemaNode(colander.String())
    permissions = colander.Sequence(colander.SchemaNode(colander.String()))
    bucket_id = colander.SchemaNode(colander.String())
    collection_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    group_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    record_id = colander.SchemaNode(colander.String(), missing=colander.drop)

    class Options:
        preserve_unknown = False


@resource.register(
    name="permissions",
    description="List of user permissions",
    plural_path="/permissions",
    object_path=None,
    plural_methods=("HEAD", "GET"),
)
class Permissions(resource.Resource):
    schema = PermissionsSchema

    def __init__(self, request, context=None):
        super().__init__(request, context)
        self.model = PermissionsModel(request)

    def _extract_sorting(self, limit):
        # Permissions entries are not stored with timestamp, so do not
        # force it.
        result = super()._extract_sorting(limit)
        without_last_modified = [s for s in result if s.field != self.model.modified_field]
        # For pagination, there must be at least one sort criteria.
        # We use ``uri`` because its values are unique.
        if "uri" not in [s.field for s in without_last_modified]:
            without_last_modified.append(Sort("uri", -1))
        return without_last_modified

    def _extract_filters(self):
        result = super()._extract_filters()
        without_last_modified = [s for s in result if s.field != self.model.modified_field]
        return without_last_modified



from kinto.core.testing import unittest

from .support import BaseWebTest


BUCKET_URL = "/buckets/blog"
GROUPS_URL = "/buckets/blog/groups"
GROUP_URL = GROUPS_URL + "/tarnac-nine"


SCHEMA = {
    "title": "Group schema",
    "type": "object",
    "required": ["phone"],
    "properties": {
        "phone": {"type": "string"},
        "members": {"type": "array", "items": {"type": "string"}},
    },
}

VALID_GROUP = {"members": ["oidc:coupat"], "phone": "+33688776655"}


class DeactivatedSchemaTest(BaseWebTest, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, "type": "Washmachine"}
        resp = self.app.put_json(
            BUCKET_URL, {"data": {"group:schema": newschema}}, headers=self.headers, status=400
        )
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json["message"])

    def test_group_are_not_invalid_if_do_not_match_schema(self):
        self.app.put_json(BUCKET_URL, {"data": {"group:schema": SCHEMA}}, headers=self.headers)
        self.app.put_json(GROUP_URL, {"data": {"phone": 42}}, headers=self.headers, status=201)


class BaseWebTestWithSchema(BaseWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_collection_schema_validation"] = "True"
        return settings


class GroupValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(
            BUCKET_URL, {"data": {"group:schema": SCHEMA}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_empty_metadata_can_be_validated(self):
        self.app.post_json(BUCKET_URL + "/groups", headers=self.headers, status=400)
        self.app.post_json(BUCKET_URL + "/groups", {"data": {}}, headers=self.headers, status=400)

    def test_groups_are_valid_if_match_schema(self):
        self.app.put_json(GROUP_URL, {"data": VALID_GROUP}, headers=self.headers, status=201)

    def test_groups_are_invalid_if_do_not_match_schema(self):
        self.app.put_json(GROUP_URL, {"data": {"phone": 42}}, headers=self.headers, status=400)


class ValidateIDField(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        schema = {"type": "object", "properties": {"id": {"type": "string", "pattern": "^[0-7]$"}}}
        self.app.put_json(BUCKET_URL, {"data": {"group:schema": schema}}, headers=self.headers)

    def test_group_id_is_accepted_if_valid(self):
        self.app.post_json(GROUPS_URL, {"data": {"id": "1"}}, headers=self.headers)

    def test_group_id_is_rejected_if_does_not_match(self):
        self.app.post_json(GROUPS_URL, {"data": {"id": "a"}}, headers=self.headers, status=400)


SCHEMA_UNRESOLVABLE = {"properties": {"phone": {"$ref": "#/definitions/phone"}}}


class CollectionUnresolvableTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(
            BUCKET_URL, {"data": {"group:schema": SCHEMA_UNRESOLVABLE}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_unresolvable_errors_handled(self):
        self.app.put_json(GROUP_URL, {"data": {"phone": 42}}, headers=self.headers, status=400)



import unittest
import uuid
from unittest import mock

import bcrypt
from pyramid.exceptions import ConfigurationError
from pyramid_mailer import get_mailer

from kinto.core import utils
from kinto.core.events import ACTIONS, ResourceChanged
from kinto.core.testing import DummyRequest, get_user_headers
from kinto.plugins.accounts import ACCOUNT_CACHE_KEY, scripts
from kinto.plugins.accounts.utils import (
    get_cached_reset_password,
    get_cached_validation_key,
    hash_password,
)
from kinto.plugins.accounts.views import on_account_created
from kinto.plugins.accounts.views.validation import on_account_activated

from .. import support


class AccountsWebTest(support.BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault("multiauth.policies", "account")
        extras.setdefault("includes", "kinto.plugins.accounts")
        extras.setdefault("account_create_principals", "system.Everyone")
        # XXX: this should be a default setting.
        extras.setdefault(
            "multiauth.policy.account.use",
            "kinto.plugins.accounts.authentication." "AccountsAuthenticationPolicy",
        )
        extras.setdefault("account_cache_ttl_seconds", "30")
        return super().get_app_settings(extras)


class AccountsValidationWebTest(AccountsWebTest):
    def setUp(self):
        self.mailer = get_mailer(self.app.app.registry)
        self.mailer.outbox = []  # Reset the outbox before each test.

    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {}
        # Enable the account validation option.
        extras.setdefault("account_validation", True)
        # Use a testing mailer.
        extras.setdefault("mail.mailer", "testing")
        # Email templates for the user creation.
        extras.setdefault(
            "account_validation.email_subject_template", "{name}, activate your account {id}"
        )
        extras.setdefault(
            "account_validation.email_body_template",
            "{activation-form-url}/{id}/{activation-key} {bad-key}",
        )
        # Email templates for the user validated confirmation.
        extras.setdefault(
            "account_validation.email_confirmation_subject_template",
            "{name}, your account {id} is now active",
        )
        extras.setdefault(
            "account_validation.email_confirmation_body_template",
            "Your account {id} has been successfully activated. Connect to {homepage}",
        )
        # Email templates for the reset password.
        extras.setdefault(
            "account_validation.email_reset_password_subject_template",
            "{name}, here is a temporary reset password for {id}",
        )
        extras.setdefault(
            "account_validation.email_reset_password_body_template",
            "You can use this temporary reset password {reset-password} to change your account {id} password",
        )
        return super().get_app_settings(extras)


class BadAccountsConfigTest(support.BaseWebTest, unittest.TestCase):
    def test_raise_configuration_if_accounts_not_mentioned(self):
        with self.assertRaises(ConfigurationError) as cm:
            self.make_app(
                {"includes": "kinto.plugins.accounts", "multiauth.policies": "basicauth"}
            )
        assert "Account policy missing" in str(cm.exception)


class HelloViewTest(AccountsWebTest):
    def test_accounts_capability_if_enabled(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("accounts", capabilities)


class HelloActivationViewTest(AccountsValidationWebTest):
    def test_account_validation_capability_if_enabled(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("accounts", capabilities)
        self.assertIn("validation_enabled", capabilities["accounts"])
        self.assertTrue(capabilities["accounts"]["validation_enabled"])


class AccountCreationTest(AccountsWebTest):
    def test_anyone_can_create_an_account(self):
        self.app.post_json("/accounts", {"data": {"id": "alice", "password": "12éé6"}}, status=201)

    def test_account_can_be_created_with_put(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "123456"}}, status=201)

    def test_password_is_stored_encrypted(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "123456"}}, status=201)
        stored = self.app.app.registry.storage.get(
            parent_id="alice", resource_name="account", object_id="alice"
        )
        assert stored["password"] != "123456"

    def test_authentication_is_accepted_if_account_exists(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        resp = self.app.get("/", headers=get_user_headers("me", "bouh"))
        assert resp.json["user"]["id"] == "account:me"

    def test_password_field_is_mandatory(self):
        self.app.post_json("/accounts", {"data": {"id": "me"}}, status=400)

    def test_id_field_is_mandatory(self):
        self.app.post_json("/accounts", {"data": {"password": "pass"}}, status=400)

    def test_id_can_be_email(self):
        self.app.put_json(
            "/accounts/alice+test@example.com", {"data": {"password": "123456"}}, status=201
        )

    def test_account_can_have_metadata(self):
        resp = self.app.post_json(
            "/accounts", {"data": {"id": "me", "password": "bouh", "age": 42}}, status=201
        )
        assert resp.json["data"]["age"] == 42

    def test_cannot_create_account_if_already_exists(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "not read"}}, status=401)

    def test_username_and_account_id_must_match(self):
        resp = self.app.put_json(
            "/accounts/alice", {"data": {"id": "bob", "password": "bouh"}}, status=400
        )
        assert "does not match" in resp.json["message"]

    def test_returns_existing_account_if_authenticated(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        self.app.post_json(
            "/accounts",
            {"data": {"id": "me", "password": "bouh"}},
            headers=get_user_headers("me", "bouh"),
            status=200,
        )

    def test_cannot_create_other_account_if_authenticated(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        resp = self.app.post_json(
            "/accounts",
            {"data": {"id": "you", "password": "bouh"}},
            headers=get_user_headers("me", "bouh"),
            status=400,
        )
        assert "do not match" in resp.json["message"]

    def test_authentication_does_not_call_bcrypt_twice(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        with mock.patch("kinto.plugins.accounts.authentication.bcrypt") as mocked_bcrypt:
            resp = self.app.get("/", headers=get_user_headers("me", "bouh"))
            assert resp.json["user"]["id"] == "account:me"

            resp = self.app.get("/", headers=get_user_headers("me", "bouh"))
            assert resp.json["user"]["id"] == "account:me"

            assert mocked_bcrypt.checkpw.call_count == 1

    def test_authentication_checks_bcrypt_again_if_password_changes(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        with mock.patch("kinto.plugins.accounts.authentication.bcrypt") as mocked_bcrypt:
            resp = self.app.get("/", headers=get_user_headers("me", "bouh"))
            assert resp.json["user"]["id"] == "account:me"

            self.app.patch_json(
                "/accounts/me",
                {"data": {"password": "blah"}},
                status=200,
                headers=get_user_headers("me", "bouh"),
            )

            resp = self.app.get("/", headers=get_user_headers("me", "blah"))
            assert resp.json["user"]["id"] == "account:me"

            assert mocked_bcrypt.checkpw.call_count == 2

    def test_authentication_refresh_the_cache_each_time_we_authenticate(self):
        hmac_secret = self.app.app.registry.settings["userid_hmac_secret"]
        cache_key = utils.hmac_digest(hmac_secret, ACCOUNT_CACHE_KEY.format("me"))

        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bouh"}}, status=201)
        resp = self.app.get("/", headers=get_user_headers("me", "bouh"))
        assert resp.json["user"]["id"] == "account:me"

        self.app.app.registry.cache.expire(cache_key, 10)

        resp = self.app.get("/", headers=get_user_headers("me", "bouh"))
        assert resp.json["user"]["id"] == "account:me"

        assert self.app.app.registry.cache.ttl(cache_key) >= 20

        resp = self.app.get("/", headers=get_user_headers("me", "blah"))
        assert "user" not in resp.json

    def test_validate_view_not_active(self):
        # The `validate` view is only active when the `account_validation` option is enabled.
        # Create the user.
        self.app.post_json(
            "/accounts", {"data": {"id": "alice@example.com", "password": "12éé6"}}, status=201
        )
        # Validate the user.
        self.app.post_json("/accounts/alice@example.com/validate/some_validation_key", status=404)

    def test_reset_password_view_not_active(self):
        # The `validate` view is only active when the `account_validation` option is enabled.
        # Create the user.
        self.app.post_json(
            "/accounts", {"data": {"id": "alice@example.com", "password": "12éé6"}}, status=201
        )
        # Ask for a reset password.
        self.app.post_json("/accounts/alice@example.com/reset-password", status=404)


class AccountValidationCreationTest(AccountsValidationWebTest):
    def test_create_account_fails_if_not_email(self):
        resp = self.app.post_json(
            "/accounts", {"data": {"id": "alice", "password": "12éé6"}}, status=400
        )
        assert "user id should match" in resp.json["message"]

    def test_create_account_stores_activated_field(self):
        uuid_string = "20e81ab7-51c0-444f-b204-f1c4cfe1aa7a"
        with mock.patch("uuid.uuid4", return_value=uuid.UUID(uuid_string)):
            resp = self.app.post_json(
                "/accounts",
                {
                    "data": {
                        "id": "alice@example.com",
                        "password": "12éé6",
                        "email-context": {
                            "name": "Alice",
                            "activation-form-url": "https://example.com",
                        },
                    }
                },
                status=201,
            )
        assert "activation-key" not in resp.json["data"]
        assert "validated" in resp.json["data"]
        assert not resp.json["data"]["validated"]
        assert len(self.mailer.outbox) == 1
        mail = self.mailer.outbox[0]  # Get the validation email.
        assert mail.sender == "admin@example.com"
        assert mail.subject == "Alice, activate your account alice@example.com"
        assert mail.recipients == ["alice@example.com"]
        # The {{bad-key}} from the template will be rendered as {bad-key} in
        # the final email, instead of failing the formatting.
        assert mail.body == f"https://example.com/alice@example.com/{uuid_string} {{bad-key}}"
        # The activation key is stored in the cache.
        assert get_cached_validation_key("alice@example.com", self.app.app.registry) == uuid_string

    def test_cant_authenticate_with_unactivated_account(self):
        self.app.post_json(
            "/accounts",
            {"data": {"id": "alice@example.com", "password": "12éé6", "activated": False}},
            status=201,
        )
        resp = self.app.get("/", headers=get_user_headers("alice@example.com", "12éé6"))
        assert "user" not in resp.json

    def test_validation_fail_bad_user(self):
        # Validation should fail on a non existing user.
        resp = self.app.post_json("/accounts/alice@example.com/validate/123", {}, status=403)
        assert "Account ID and activation key do not match" in resp.json["message"]

    def test_validation_fail_bad_activation_key(self):
        uuid_string = "20e81ab7-51c0-444f-b204-f1c4cfe1aa7a"
        with mock.patch("uuid.uuid4", return_value=uuid.UUID(uuid_string)):
            self.app.post_json(
                "/accounts", {"data": {"id": "alice@example.com", "password": "12éé6"}}, status=201
            )
        # Validate the user.
        resp = self.app.post_json(
            "/accounts/alice@example.com/validate/bad-activation-key", {}, status=403
        )
        assert "Account ID and activation key do not match" in resp.json["message"]
        # The activation key is still in the cache
        assert get_cached_validation_key("alice@example.com", self.app.app.registry) is not None

    def test_validation_validates_user(self):
        # On user activation the 'validated' field is set to True.
        uuid_string = "20e81ab7-51c0-444f-b204-f1c4cfe1aa7a"
        with mock.patch("uuid.uuid4", return_value=uuid.UUID(uuid_string)):
            self.app.post_json(
                "/accounts",
                {
                    "data": {
                        "id": "alice@example.com",
                        "password": "12éé6",
                        "email-context": {"name": "Alice", "homepage": "https://example.com"},
                    }
                },
                status=201,
            )
        resp = self.app.post_json(
            "/accounts/alice@example.com/validate/" + uuid_string, {}, status=200
        )
        assert "validated" in resp.json
        assert resp.json["validated"]
        # An active user can authenticate.
        resp = self.app.get("/", headers=get_user_headers("alice@example.com", "12éé6"))
        assert resp.json["user"]["id"] == "account:alice@example.com"
        # Once activated, the activation key is removed from the cache.
        assert get_cached_validation_key("alice@example.com", self.app.app.registry) is None
        assert len(self.mailer.outbox) == 2  # Validation email, reset password email.
        mail = self.mailer.outbox[1]  # Get the confirmation email.
        assert mail.sender == "admin@example.com"
        assert mail.subject == "Alice, your account alice@example.com is now active"
        assert mail.recipients == ["alice@example.com"]
        assert (
            mail.body
            == "Your account alice@example.com has been successfully activated. Connect to https://example.com"
        )

    def test_previously_created_accounts_can_still_authenticate(self):
        """Accounts created before activating the 'account validation' option can still authenticate."""
        # Create an account without going through the accounts API.
        hashed_password = hash_password("12éé6")
        self.app.app.registry.storage.create(
            parent_id="alice",
            resource_name="account",
            record={"id": "alice", "password": hashed_password},
        )
        resp = self.app.get("/", headers=get_user_headers("alice", "12éé6"))
        assert resp.json["user"]["id"] == "account:alice"

    def test_reset_password_bad_user(self):
        resp = self.app.post_json("/accounts/alice@example.com/reset-password", {}, status=200)
        # Don't give information on the existence of a user id: return a generic message.
        assert resp.json["message"] == "A temporary reset password has been sent by mail"
        # Make sure no email was sent.
        assert len(self.mailer.outbox) == 0

    def test_reset_password_bad_email(self):
        # Create an account without going through the accounts API.
        hashed_password = hash_password("12éé6")
        self.app.app.registry.storage.create(
            parent_id="alice",
            resource_name="account",
            record={"id": "alice", "password": hashed_password},
        )
        resp = self.app.post_json("/accounts/alice/reset-password", {}, status=400)
        assert "user id should match" in resp.json["message"]

    def test_reset_password_sends_email(self):
        reset_password = "20e81ab7-51c0-444f-b204-f1c4cfe1aa7a"
        with mock.patch("uuid.uuid4", return_value=uuid.UUID(reset_password)):
            # Create the user.
            self.app.post_json(
                "/accounts", {"data": {"id": "alice@example.com", "password": "12éé6"}}, status=201
            )
            # Ask for a reset password.
            resp = self.app.post_json(
                "/accounts/alice@example.com/reset-password",
                {"data": {"email-context": {"name": "Alice"}}},
                status=200,
            )
        assert resp.json["message"] == "A temporary reset password has been sent by mail"
        assert len(self.mailer.outbox) == 2  # Validation email, reset password email.
        mail = self.mailer.outbox[1]  # Get the reset password email
        assert mail.sender == "admin@example.com"
        assert mail.subject == "Alice, here is a temporary reset password for alice@example.com"
        assert (
            mail.body
            == f"You can use this temporary reset password {reset_password} to change your account alice@example.com password"
        )
        # The reset password is stored in the cache.
        cached_password = get_cached_reset_password(
            "alice@example.com", self.app.app.registry
        ).encode(encoding="utf-8")
        pwd_str = reset_password.encode(encoding="utf-8")
        assert bcrypt.checkpw(pwd_str, cached_password)

    def test_fail_use_reset_password_bad_data(self):
        validation_key = reset_password = "20e81ab7-51c0-444f-b204-f1c4cfe1aa7a"
        with mock.patch("uuid.uuid4", return_value=uuid.UUID(reset_password)):
            # Create the user.
            self.app.post_json(
                "/accounts", {"data": {"id": "alice@example.com", "password": "12éé6"}}, status=201
            )
            # Validate the user.
            resp = self.app.post_json(
                "/accounts/alice@example.com/validate/" + validation_key, status=200
            )
            # Ask for a reset password.
            self.app.post_json("/accounts/alice@example.com/reset-password", status=200)
        # Using reset password needs data.
        self.app.put_json(
            "/accounts/alice@example.com",
            headers=get_user_headers("alice@example.com", reset_password),
            status=401,
        )
        # Using reset password needs password field.
        self.app.put_json(
            "/accounts/alice@example.com",
            {"data": {"foo": "bar"}},
            headers=get_user_headers("alice@example.com", reset_password),
            status=401,
        )
        # Using reset password accepts only password field.
        self.app.put_json(
            "/accounts/alice@example.com",
            {"data": {"password": "newpass", "foo": "bar"}},
            headers=get_user_headers("alice@example.com", reset_password),
            status=401,
        )
        # Using wrong reset password fails.
        self.app.put_json(
            "/accounts/alice@example.com",
            {"data": {"password": "newpass"}},
            headers=get_user_headers("alice@example.com", "some random password"),
            status=401,
        )
        # Can't use the reset password to modify other resources than accounts.
        resp = self.app.post_json(
            "/buckets/default/collections",
            {"data": {"id": "some_collection_id"}},
            headers=get_user_headers("alice@example.com", reset_password),
            status=401,
        )
        assert resp.json["message"] == "Please authenticate yourself to use this endpoint."
        # Can't use reset password to authenticate.
        resp = self.app.get("/", headers=get_user_headers("alice@example.com", reset_password))
        assert "user" not in resp.json

    def test_use_reset_password_to_change_password(self):
        validation_key = reset_password = "20e81ab7-51c0-444f-b204-f1c4cfe1aa7a"
        with mock.patch("uuid.uuid4", return_value=uuid.UUID(reset_password)):
            # Create the user.
            self.app.post_json(
                "/accounts", {"data": {"id": "alice@example.com", "password": "12éé6"}}, status=201
            )
            # Validate the user.
            resp = self.app.post_json(
                "/accounts/alice@example.com/validate/" + validation_key, {}, status=200
            )
            # Ask for a reset password.
            self.app.post_json("/accounts/alice@example.com/reset-password", {}, status=200)
        # Use reset password to set a new password.
        self.app.patch_json(
            "/accounts/alice@example.com",
            {"data": {"password": "newpass"}},
            headers=get_user_headers("alice@example.com", reset_password),
            status=200,
        )
        # Can use the new password to authenticate.
        resp = self.app.get("/", headers=get_user_headers("alice@example.com", "newpass"))
        assert resp.json["user"]["id"] == "account:alice@example.com"
        # The user hasn't changed.
        resp = self.app.get(
            "/accounts/alice@example.com", headers=get_user_headers("alice@example.com", "newpass")
        )
        assert resp.json["data"]["id"] == "alice@example.com"
        assert resp.json["data"]["validated"]
        # The reset password isn't in the cache anymore
        assert get_cached_reset_password("alice@example.com", self.app.app.registry) is None
        # Can't use the reset password anymore to authenticate.
        resp = self.app.get("/", headers=get_user_headers("alice@example.com", reset_password))
        assert "user" not in resp.json

    def test_user_creation_listener(self):
        request = DummyRequest()
        impacted_object = {"new": {"id": "alice", "password": "12éé6"}}
        with mock.patch("kinto.plugins.accounts.mails.Emailer.send_mail") as mocked_send_mail:
            # No email sent if account validation is not enabled.
            event = ResourceChanged({"action": ACTIONS.UPDATE.value}, [impacted_object], request)
            on_account_created(event)
            mocked_send_mail.assert_not_called()
            # No email sent if there's no activation key in the cache.
            request.registry.settings["account_validation"] = True
            event = ResourceChanged({"action": ACTIONS.UPDATE.value}, [impacted_object], request)
            request.registry.cache.get = mock.MagicMock(return_value=None)
            on_account_created(event)
            mocked_send_mail.assert_not_called()
            # Email sent if there is an activation key in the cache.
            request.registry.cache.get = mock.MagicMock(return_value="some activation key")
            on_account_created(event)
            mocked_send_mail.assert_called_once()

    def test_user_validation_listener(self):
        request = DummyRequest()
        old_inactive = {"id": "alice", "password": "12éé6", "validated": False}
        old_active = {"id": "alice", "password": "12éé6", "validated": True}
        new_inactive = {"id": "alice", "password": "12éé6", "validated": False}
        new_active = {"id": "alice", "password": "12éé6", "validated": True}
        with mock.patch("kinto.plugins.accounts.mails.Emailer.send_mail") as mocked_send_mail:
            # No email sent if account validation is not enabled.
            event = ResourceChanged(
                {"action": ACTIONS.UPDATE.value},
                [{"old": old_inactive, "new": new_inactive}],
                request,
            )
            on_account_activated(event)
            mocked_send_mail.assert_not_called()
            # No email sent if the old account was already active.
            request.registry.settings["account_validation"] = True
            event = ResourceChanged(
                {"action": ACTIONS.UPDATE.value}, [{"old": old_active, "new": new_active}], request
            )
            request.registry.cache.get = mock.MagicMock(return_value=None)
            on_account_activated(event)
            mocked_send_mail.assert_not_called()
            # No email sent if the new account is still inactive.
            event = ResourceChanged(
                {"action": ACTIONS.UPDATE.value},
                [{"old": old_inactive, "new": new_inactive}],
                request,
            )
            request.registry.cache.get = mock.MagicMock(return_value=None)
            on_account_activated(event)
            mocked_send_mail.assert_not_called()
            # Email sent if there is an activation key in the cache.
            event = ResourceChanged(
                {"action": ACTIONS.UPDATE.value},
                [{"old": old_inactive, "new": new_active}],
                request,
            )
            on_account_activated(event)
            mocked_send_mail.assert_called_once()


class AccountUpdateTest(AccountsWebTest):
    def setUp(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "123456"}}, status=201)

    def test_password_can_be_changed(self):
        self.app.put_json(
            "/accounts/alice",
            {"data": {"password": "bouh"}},
            headers=get_user_headers("alice", "123456"),
            status=200,
        )

    def test_authentication_with_old_password_is_denied_after_change(self):
        self.app.put_json(
            "/accounts/alice",
            {"data": {"password": "bouh"}},
            headers=get_user_headers("alice", "123456"),
            status=200,
        )
        self.app.get("/accounts/alice", headers=get_user_headers("alice", "123456"), status=401)

    def test_authentication_with_new_password_is_accepted_after_change(self):
        self.app.put_json(
            "/accounts/alice",
            {"data": {"password": "bouh"}},
            headers=get_user_headers("alice", "123456"),
            status=200,
        )
        self.app.get("/accounts/alice", headers=get_user_headers("alice", "bouh"))

    def test_username_and_account_id_must_match(self):
        resp = self.app.patch_json(
            "/accounts/alice",
            {"data": {"id": "bob", "password": "bouh"}},
            headers=get_user_headers("alice", "123456"),
            status=400,
        )
        assert "does not match" in resp.json["message"]

    def test_cannot_patch_unknown_account(self):
        self.app.patch_json(
            "/accounts/bob",
            {"data": {"password": "bouh"}},
            headers=get_user_headers("alice", "123456"),
            status=403,
        )

    def test_cannot_patch_someone_else_account(self):
        self.app.put_json("/accounts/bob", {"data": {"password": "bob"}}, status=201)
        self.app.patch_json(
            "/accounts/bob",
            {"data": {"password": "bouh"}},
            headers=get_user_headers("alice", "123456"),
            status=403,
        )

    def test_metadata_can_be_changed(self):
        resp = self.app.patch_json(
            "/accounts/alice",
            {"data": {"age": "captain"}},
            headers=get_user_headers("alice", "123456"),
        )
        assert resp.json["data"]["age"] == "captain"

    def test_changing_metadata_does_not_change_password(self):
        headers = get_user_headers("alice", "123456")
        url = "/accounts/alice"
        resp = self.app.get(url, headers=headers)
        before = resp.json["data"]["password"]

        self.app.patch_json(url, {"data": {"age": "captain"}}, headers=headers)

        resp = self.app.get(url, headers=headers)
        after = resp.json["data"]["password"]
        assert before == after


class AccountDeleteTest(AccountsWebTest):
    def setUp(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "123456"}}, status=201)

    def test_account_can_be_deleted(self):
        self.app.delete("/accounts/alice", headers=get_user_headers("alice", "123456"))

    def test_authentication_is_denied_after_delete(self):
        self.app.delete("/accounts/alice", headers=get_user_headers("alice", "123456"))
        self.app.get("/accounts/alice", headers=get_user_headers("alice", "123456"), status=401)


class AccountViewsTest(AccountsWebTest):
    def setUp(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "123456"}}, status=201)
        self.app.put_json("/accounts/bob", {"data": {"password": "azerty"}}, status=201)

    def test_account_list_is_forbidden_if_anonymous(self):
        self.app.get("/accounts", status=401)

    def test_account_detail_is_forbidden_if_anonymous(self):
        self.app.get("/accounts/alice", status=401)

    def test_accounts_list_contains_only_one_record(self):
        resp = self.app.get("/accounts", headers=get_user_headers("alice", "123456"))
        assert len(resp.json["data"]) == 1

    def test_account_record_can_be_obtained_if_authenticated(self):
        self.app.get("/accounts/alice", headers=get_user_headers("alice", "123456"))

    def test_cannot_obtain_someone_else_account(self):
        self.app.get("/accounts/bob", headers=get_user_headers("alice", "123456"), status=403)

    def test_cannot_obtain_unknown_account(self):
        self.app.get("/accounts/jeanine", headers=get_user_headers("alice", "123456"), status=403)


class PermissionsEndpointTest(AccountsWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_permissions_endpoint"] = "True"
        return settings

    def setUp(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "123456"}}, status=201)
        self.headers = get_user_headers("alice", "123456")
        self.app.put("/buckets/a", headers=self.headers)
        self.app.put("/buckets/a/collections/b", headers=self.headers)
        self.app.put("/buckets/a/collections/b/records/c", headers=self.headers)

    def test_permissions_endpoint_is_compatible_with_accounts_plugin(self):
        resp = self.app.get("/permissions", headers=self.headers)
        uris = [r["uri"] for r in resp.json["data"]]
        assert uris == [
            "/buckets/a/collections/b/records/c",
            "/buckets/a/collections/b",
            "/buckets/a",
            "/accounts/alice",
            "/",
        ]

    def test_account_create_read_write_permissions_(self):
        resp = self.app.get("/permissions", headers=self.headers)
        buckets = resp.json["data"]
        account = buckets[3]
        account["permissions"].sort()
        root = buckets[4]
        root["permissions"].sort()
        self.assertEqual(
            account,
            {
                "account_id": "alice",
                "id": "alice",
                "permissions": ["read", "write"],
                "resource_name": "account",
                "uri": "/accounts/alice",
            },
        )
        self.assertEqual(
            root,
            {
                "permissions": ["account:create", "bucket:create"],
                "resource_name": "root",
                "uri": "/",
            },
        )


class PermissionsEndpointTestUnauthenticatedCreatePermission(AccountsWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_permissions_endpoint"] = "True"
        settings["bucket_create_principals"] = "system.Everyone"
        return settings

    def setUp(self):
        self.everyone_headers = get_user_headers("")

    def test_permissions_endpoint_is_compatible_with_accounts_plugin(self):
        resp = self.app.get("/permissions", headers=self.everyone_headers)
        buckets = resp.json["data"]
        root = buckets[0]
        root["permissions"].sort()
        self.assertEqual(
            root,
            {
                "permissions": ["account:create", "bucket:create"],
                "resource_name": "root",
                "uri": "/",
            },
        )


class AdminTest(AccountsWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["account_write_principals"] = "account:admin"
        settings["account_read_principals"] = "account:admin"
        return settings

    def setUp(self):
        self.app.put_json("/accounts/admin", {"data": {"password": "123456"}}, status=201)
        self.admin_headers = get_user_headers("admin", "123456")

        self.app.put_json("/accounts/bob", {"data": {"password": "987654"}}, status=201)

    def test_admin_can_get_other_accounts(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "azerty"}}, status=201)

        resp = self.app.get("/accounts/alice", headers=get_user_headers("alice", "azerty"))
        assert resp.json["data"]["id"] == "alice"

    def test_admin_can_delete_other_accounts(self):
        self.app.delete_json("/accounts/bob", headers=self.admin_headers)
        self.app.get("/accounts/bob", headers=get_user_headers("bob", "987654"), status=401)

    def test_admin_can_delete_all_accounts(self):
        self.app.delete_json("/accounts", headers=self.admin_headers)

        self.app.get("/accounts/bob", headers=get_user_headers("bob", "987654"), status=401)

    def test_admin_can_set_others_password(self):
        self.app.patch_json(
            "/accounts/bob", {"data": {"password": "bouh"}}, headers=self.admin_headers
        )
        self.app.get("/accounts/bob", headers=get_user_headers("bob", "987654"), status=401)
        self.app.get("/accounts/bob", headers=get_user_headers("bob", "bouh"))

    def test_admin_can_retrieve_accounts_list(self):
        self.app.put_json("/accounts/alice", {"data": {"password": "azerty"}}, status=201)

        resp = self.app.get("/accounts", headers=self.admin_headers)
        usernames = [r["id"] for r in resp.json["data"]]
        assert sorted(usernames) == ["admin", "alice", "bob"]

    def test_user_created_by_admin_with_put_can_see_her_record(self):
        self.app.put_json(
            "/accounts/alice", {"data": {"password": "bouh"}}, headers=self.admin_headers
        )

        resp = self.app.get("/accounts/alice", headers=get_user_headers("alice", "bouh"))
        assert resp.json["permissions"] == {"write": ["account:alice"]}

    def test_user_created_by_admin_with_post_can_see_her_record(self):
        self.app.post_json(
            "/accounts", {"data": {"id": "alice", "password": "bouh"}}, headers=self.admin_headers
        )

        resp = self.app.get("/accounts/alice", headers=get_user_headers("alice", "bouh"))
        assert resp.json["permissions"] == {"write": ["account:alice"]}

    def test_user_created_by_admin_can_delete_her_record(self):
        self.app.post_json(
            "/accounts", {"data": {"id": "alice", "password": "bouh"}}, headers=self.admin_headers
        )

        self.app.delete("/accounts/alice", headers=get_user_headers("alice", "bouh"))


class CheckAdminCreateTest(AccountsWebTest):
    def test_raise_if_create_but_no_write(self):
        with self.assertRaises(ConfigurationError):
            self.make_app({"account_create_principals": "account:admin"})


class CreateOtherUserTest(AccountsWebTest):
    def setUp(self):
        self.app.put_json("/accounts/bob", {"data": {"password": "123456"}}, status=201)
        self.bob_headers = get_user_headers("bob", "123456")

    def test_create_other_id_is_still_required(self):
        self.app.post_json(
            "/accounts", {"data": {"password": "azerty"}}, status=400, headers=self.bob_headers
        )

    def test_create_other_forbidden_without_write(self):
        self.app.put_json(
            "/accounts/alice",
            {"data": {"password": "azerty"}},
            status=400,
            headers=self.bob_headers,
        )


class WithBasicAuthTest(AccountsWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {"multiauth.policies": "account basicauth"}
        return super().get_app_settings(extras)

    def test_password_field_is_mandatory(self):
        self.app.post_json("/accounts", {"data": {"id": "me"}}, status=400)

    def test_id_field_is_mandatory(self):
        self.app.post_json("/accounts", {"data": {"password": "pass"}}, status=400)

    def test_fallsback_on_basicauth(self):
        self.app.post_json("/accounts", {"data": {"id": "me", "password": "bleh"}})

        resp = self.app.get("/", headers=get_user_headers("me", "wrong"))
        assert "basicauth" in resp.json["user"]["id"]

        resp = self.app.get("/", headers=get_user_headers("me", "bleh"))
        assert "account" in resp.json["user"]["id"]

    def test_raise_configuration_if_wrong_error(self):
        with self.assertRaises(ConfigurationError):
            self.make_app({"multiauth.policies": "basicauth account"})


class CreateUserTest(unittest.TestCase):
    def setUp(self):
        self.registry = mock.MagicMock()
        self.registry.settings = {"includes": "kinto.plugins.accounts"}

    def test_create_user_in_read_only_displays_an_error(self):
        with mock.patch("kinto.plugins.accounts.scripts.logger") as mocked:
            self.registry.settings["readonly"] = "true"
            code = scripts.create_user({"registry": self.registry})
            assert code == 51
            mocked.error.assert_called_once_with("Cannot create a user with " "a readonly server.")

    def test_create_user_when_not_included_displays_an_error(self):
        with mock.patch("kinto.plugins.accounts.scripts.logger") as mocked:
            self.registry.settings["includes"] = ""
            code = scripts.create_user({"registry": self.registry})
            assert code == 52
            mocked.error.assert_called_once_with(
                "Cannot create a user when the accounts " "plugin is not installed."
            )

    def test_create_user_with_an_invalid_username_and_password_confirmation_recovers(self):
        with mock.patch("kinto.plugins.accounts.scripts.input", side_effect=["&zert", "username"]):
            with mock.patch(
                "kinto.plugins.accounts.scripts.getpass.getpass",
                side_effect=["password", "p4ssw0rd", "password", "password"],
            ):
                code = scripts.create_user({"registry": self.registry}, None, None)
                assert self.registry.storage.update.call_count == 1
                self.registry.permission.add_principal_to_ace.assert_called_with(
                    "/accounts/username", "write", "account:username"
                )
                assert code == 0

    def test_create_user_with_a_valid_username_and_password_confirmation(self):
        with mock.patch("kinto.plugins.accounts.scripts.input", return_value="username"):
            with mock.patch(
                "kinto.plugins.accounts.scripts.getpass.getpass", return_value="password"
            ):
                code = scripts.create_user({"registry": self.registry})
                assert self.registry.storage.update.call_count == 1
                self.registry.permission.add_principal_to_ace.assert_called_with(
                    "/accounts/username", "write", "account:username"
                )
                assert code == 0

    def test_create_user_with_valid_username_and_password_parameters(self):
        code = scripts.create_user({"registry": self.registry}, "username", "password")
        assert self.registry.storage.update.call_count == 1
        self.registry.permission.add_principal_to_ace.assert_called_with(
            "/accounts/username", "write", "account:username"
        )
        assert code == 0

    def test_create_user_aborted_by_eof(self):
        with mock.patch("kinto.plugins.accounts.scripts.input", side_effect=EOFError):
            code = scripts.create_user({"registry": self.registry})
            assert code == 53



from kinto import DEFAULT_SETTINGS
from kinto import main as kinto_main
from kinto.core import testing


MINIMALIST_BUCKET = {}
MINIMALIST_COLLECTION = {}
MINIMALIST_GROUP = {"data": dict(members=["fxa:user"])}
MINIMALIST_RECORD = {"data": dict(name="Hulled Barley", type="Whole Grain")}
USER_PRINCIPAL = "basicauth:8a931a10fc88ab2f6d1cc02a07d3a81b5d4768f6f13e85c5" "d8d4180419acb1b4"


class BaseWebTest(testing.BaseWebTest):
    api_prefix = "v1"
    entry_point = kinto_main
    principal = USER_PRINCIPAL

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.headers.update(testing.get_user_headers("mat"))

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = {**DEFAULT_SETTINGS, "multiauth.policies": "basicauth"}
        if extras is not None:
            settings.update(extras)
        settings = super().get_app_settings(extras=settings)
        return settings

    def create_group(self, bucket_id, group_id, members=None):
        if members is None:
            group = MINIMALIST_GROUP
        else:
            group = {"data": {"members": members}}
        group_url = "/buckets/{}/groups/{}".format(bucket_id, group_id)
        self.app.put_json(group_url, group, headers=self.headers, status=201)

    def create_bucket(self, bucket_id):
        self.app.put_json(
            "/buckets/{}".format(bucket_id), MINIMALIST_BUCKET, headers=self.headers, status=201
        )



import logging

from pyramid import httpexceptions
from pyramid.authorization import Authenticated
from pyramid.httpexceptions import HTTPTemporaryRedirect
from pyramid.security import NO_PERMISSION_REQUIRED, forget
from pyramid.settings import asbool
from pyramid.view import view_config

from kinto.core.errors import ERRORS, http_error, request_GET
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.utils import reapply_cors


logger = logging.getLogger()


@view_config(context=httpexceptions.HTTPForbidden, permission=NO_PERMISSION_REQUIRED)
def authorization_required(response, request):
    """Distinguish authentication required (``401 Unauthorized``) from
    not allowed (``403 Forbidden``).
    """
    if Authenticated not in request.effective_principals:
        if response.content_type != "application/json":
            error_msg = "Please authenticate yourself to use this endpoint."
            response = http_error(
                httpexceptions.HTTPUnauthorized(),
                errno=ERRORS.MISSING_AUTH_TOKEN,
                message=error_msg,
            )
        response.headers.extend(forget(request))
        return response

    if response.content_type != "application/json":
        error_msg = "This user cannot access this resource."
        response = http_error(
            httpexceptions.HTTPForbidden(), errno=ERRORS.FORBIDDEN, message=error_msg
        )
    return reapply_cors(request, response)


@view_config(context=httpexceptions.HTTPNotFound, permission=NO_PERMISSION_REQUIRED)
def page_not_found(response, request):
    """Return a JSON 404 error response."""
    config_key = "trailing_slash_redirect_enabled"
    redirect_enabled = request.registry.settings[config_key]
    trailing_slash_redirection_enabled = asbool(redirect_enabled)

    querystring = request.url[(request.url.rindex(request.path) + len(request.path)) :]

    errno = ERRORS.MISSING_RESOURCE
    error_msg = "The resource you are looking for could not be found."

    if not request.path.startswith(f"/{request.registry.route_prefix}"):
        errno = ERRORS.VERSION_NOT_AVAILABLE
        error_msg = "The requested API version is not available " "on this server."
    elif trailing_slash_redirection_enabled:
        redirect = None

        if request.path.endswith("/"):
            path = request.path.rstrip("/")
            redirect = f"{path}{querystring}"
        elif request.path == f"/{request.registry.route_prefix}":
            # Case for /v0 -> /v0/
            redirect = f"/{request.registry.route_prefix}/{querystring}"

        if redirect:
            resp = HTTPTemporaryRedirect(redirect)
            cache_seconds = int(request.registry.settings["trailing_slash_redirect_ttl_seconds"])
            if cache_seconds >= 0:
                resp.cache_expires(cache_seconds)
            return reapply_cors(request, resp)

    if response.content_type != "application/json":
        response = http_error(httpexceptions.HTTPNotFound(), errno=errno, message=error_msg)
    return reapply_cors(request, response)


@view_config(context=httpexceptions.HTTPServiceUnavailable, permission=NO_PERMISSION_REQUIRED)
def service_unavailable(response, request):
    if response.content_type != "application/json":
        error_msg = (
            "Service temporary unavailable "
            "due to overloading or maintenance, please retry later."
        )
        response = http_error(response, errno=ERRORS.BACKEND, message=error_msg)

    retry_after = request.registry.settings["retry_after_seconds"]
    response.headers["Retry-After"] = str(retry_after)
    return reapply_cors(request, response)


@view_config(context=httpexceptions.HTTPMethodNotAllowed, permission=NO_PERMISSION_REQUIRED)
def method_not_allowed(context, request):
    if context.content_type == "application/json":
        return context

    response = http_error(
        context, errno=ERRORS.METHOD_NOT_ALLOWED, message="Method not allowed on this endpoint."
    )
    return reapply_cors(request, response)


@view_config(context=Exception, permission=NO_PERMISSION_REQUIRED)
@view_config(context=httpexceptions.HTTPException, permission=NO_PERMISSION_REQUIRED)
def error(context, request):
    """Catch server errors and trace them."""
    if isinstance(context, httpexceptions.Response):
        return reapply_cors(request, context)

    if isinstance(context, storage_exceptions.IntegrityError):
        error_msg = "Integrity constraint violated, please retry."
        response = http_error(
            httpexceptions.HTTPConflict(), errno=ERRORS.CONSTRAINT_VIOLATED, message=error_msg
        )
        retry_after = request.registry.settings["retry_after_seconds"]
        response.headers["Retry-After"] = str(retry_after)
        return reapply_cors(request, response)

    # Log some information about current request.
    extra = {"path": request.path, "method": request.method}
    qs = dict(request_GET(request))
    if qs:
        extra["querystring"] = qs
    # Take errno from original exception, or undefined if unknown/unhandled.
    try:
        extra["errno"] = context.errno.value
    except AttributeError:
        extra["errno"] = ERRORS.UNDEFINED.value

    if isinstance(context, storage_exceptions.BackendError):
        logger.critical(context.original, extra=extra, exc_info=context)
        response = httpexceptions.HTTPServiceUnavailable()
        return service_unavailable(response, request)

    # Within the exception view, sys.exc_info() will return null.
    # see https://github.com/python/cpython/blob/ce9e62544/Lib/logging/__init__.py#L1460-L1462
    logger.error(context, extra=extra, exc_info=context)

    error_msg = "A programmatic error occured, developers have been informed."
    info = request.registry.settings["error_info_link"]
    response = http_error(httpexceptions.HTTPInternalServerError(), message=error_msg, info=info)

    return reapply_cors(request, response)



import os.path
import random
import re
import unittest
import uuid
from urllib.parse import urljoin

import requests

from kinto import HTTP_API_VERSION


# abspath here because __file__ may be relative if it is in __main__
__HERE__ = os.path.abspath(os.path.dirname(__file__))

SERVER_URL = "http://localhost:8888/v1"
DEFAULT_AUTH = ("user", "p4ssw0rd")


def build_task():
    suffix = str(uuid.uuid4())
    data = {
        "description": "Task description {0}".format(suffix),
        "status": random.choice(("todo", "doing")),
    }
    return data


class FunctionalTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(FunctionalTest, self).__init__(*args, **kwargs)
        # XXX Read the configuration from env variables.
        self.server_url = SERVER_URL
        self.auth = DEFAULT_AUTH
        self.session = requests.Session()
        self.session.auth = DEFAULT_AUTH

    def setUp(self):
        # Create accounts used in the different tests.
        requests.post(
            urljoin(self.server_url, "/accounts"),
            json={"data": {"id": "user", "password": "p4ssw0rd"}},
        )
        requests.post(
            urljoin(self.server_url, "/accounts"),
            json={"data": {"id": "bob", "password": "s3cr3t"}},
        )
        requests.post(
            urljoin(self.server_url, "/accounts"),
            json={"data": {"id": "alice", "password": "wh1sp3r"}},
        )
        requests.post(
            urljoin(self.server_url, "/accounts"),
            json={"data": {"id": "mary", "password": "s4f3"}},
        )

    def tearDown(self):
        # Delete all the created objects
        flush_url = urljoin(self.server_url, "/__flush__")
        resp = requests.post(flush_url)
        resp.raise_for_status()

    def test_http_api_version(self):
        resp = self.session.get(urljoin(self.server_url, "/v1"))
        resp.raise_for_status()
        self.assertEqual(resp.json()["http_api_version"], HTTP_API_VERSION)

    def test_user_default_bucket_tutorial(self):
        collection_id = "tasks-%s" % uuid.uuid4()
        collection_url = urljoin(
            self.server_url, "/buckets/default/collections/{}/records".format(collection_id)
        )
        task = build_task()

        # Create a task on a default collection
        # 201 created with data and permission.write
        resp = self.session.post(collection_url, json={"data": task})
        resp.raise_for_status()
        self.assertEqual(resp.status_code, 201)
        record = resp.json()

        self.assertEqual(record["data"]["description"], task["description"])
        self.assertEqual(record["data"]["status"], task["status"])
        self.assertIn("write", record["permissions"])

        # Create a new one with PUT and If-None-Match: "*"
        task = build_task()
        record_id = str(uuid.uuid4())
        record_url = urljoin(self.server_url, "{}/{}".format(collection_url, record_id))
        resp = self.session.put(record_url, json={"data": task}, headers={"If-None-Match": "*"})
        resp.raise_for_status()
        resp.raise_for_status()
        self.assertEqual(resp.status_code, 201)
        record = resp.json()

        self.assertEqual(record["data"]["description"], task["description"])
        self.assertEqual(record["data"]["status"], task["status"])
        self.assertIn("write", record["permissions"])

        # Fetch the collection list and see the tasks (save the etag)
        resp = self.session.get(collection_url)
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record["data"]), 2)
        etag = resp.headers["ETag"]

        # Fetch the collection from the Etag and see nothing new
        resp = self.session.get(collection_url, params={"_since": etag.strip('"')})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record["data"]), 0)

        # Update a task
        resp = self.session.patch(record_url, json={"data": {"status": "done"}})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(record["data"]["status"], "done")
        self.assertIn("write", record["permissions"])

        # Try an update with If-Match on the saved ETag and see it fails
        resp = self.session.patch(
            record_url, json={"data": {"status": "done"}}, headers={"If-Match": etag}
        )
        self.assertEqual(resp.status_code, 412)
        self.assertEqual(resp.headers["ETag"], '"%s"' % record["data"]["last_modified"])

        # Get the list of records and update the ETag
        resp = self.session.get("%s?_since=%s" % (collection_url, etag.strip('"')))
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record["data"]), 1)
        etag = resp.headers["ETag"]

        # Try an update with If-Match on the new ETag and see it works
        resp = self.session.patch(
            record_url, json={"data": {"status": "done"}}, headers={"If-Match": etag}
        )
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        etag = resp.headers["ETag"]

        # Delete the record with If-Match
        resp = self.session.delete(record_url, headers={"If-Match": etag})

        self.assertEqual(resp.status_code, 200)

        # Try the collection get with the ``_since`` parameter
        resp = self.session.get(
            "%s?_since=%s" % (collection_url, etag.strip('"')), headers={"If-None-Match": etag}
        )
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record["data"]), 1)
        self.assertIn("deleted", record["data"][0])

        # Delete all the things
        resp = self.session.delete(collection_url)
        self.assertEqual(resp.status_code, 200)

    def test_user_shared_bucket_tutorial(self):
        bucket_id = "bucket-%s" % uuid.uuid4()
        collection_id = "tasks-%s" % uuid.uuid4()
        bucket_url = urljoin(self.server_url, "/buckets/{}".format(bucket_id))
        collection_url = "{}/collections/{}/records".format(bucket_url, collection_id)

        # Create a new bucket and check for permissions
        resp = self.session.put(bucket_url)
        # In case of concurrent execution, it may have been created already.
        self.assertIn(resp.status_code, (200, 201))
        bucket = resp.json()
        self.assertIn("write", bucket["permissions"])

        # Create a new collection and check for permissions
        permissions = {"record:create": ["system.Authenticated"]}
        resp = self.session.put(
            re.sub("/records$", "", collection_url), json={"permissions": permissions}
        )
        # In case of concurrent execution, it may have been created already.
        self.assertIn(resp.status_code, (200, 201))
        collection = resp.json()
        self.assertIn("record:create", collection["permissions"])
        self.assertIn("system.Authenticated", collection["permissions"]["record:create"])

        # Create a new tasks for Alice
        alice_auth = ("alice", "wh1sp3r")
        alice_task = build_task()
        resp = self.session.post(collection_url, json={"data": alice_task}, auth=alice_auth)
        self.assertEqual(resp.status_code, 201)
        collection = resp.json()
        self.assertIn("write", collection["permissions"])
        alice_task_id = collection["data"]["id"]

        bob_auth = ("bob", "s3cr3t")

        # Bob has no task yet.
        resp = self.session.get(collection_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()["data"]), 0)

        # Create a new tasks for Bob
        bob_task = build_task()
        resp = self.session.post(collection_url, json={"data": bob_task}, auth=bob_auth)
        self.assertEqual(resp.status_code, 201)
        collection = resp.json()
        self.assertIn("write", collection["permissions"])
        bob_user_id = collection["permissions"]["write"][0]
        bob_task_id = collection["data"]["id"]

        # Now that he has a task, he should see his.
        resp = self.session.get(collection_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)
        records = resp.json()["data"]
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]["id"], bob_task_id)

        record_url = "{}/{}".format(collection_url, alice_task_id)

        # Share Alice's task with Bob
        resp = self.session.patch(
            record_url, json={"permissions": {"read": [bob_user_id]}}, auth=alice_auth
        )
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertIn("write", record["permissions"])
        alice_task_id = record["data"]["id"]

        # Check that Bob can access it
        resp = self.session.get(record_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)

        # Get mary's userid
        mary_auth = ("mary", "s4f3")
        resp = self.session.get("{}/".format(self.server_url), auth=mary_auth)
        self.assertEqual(resp.status_code, 200)
        hello = resp.json()
        mary_user_id = hello["user"]["id"]

        # Allow group creation on bucket
        permissions = {"group:create": ["system.Authenticated"]}
        resp = self.session.put(bucket_url, json={"permissions": permissions})
        self.assertEqual(resp.status_code, 200)
        bucket = resp.json()
        self.assertIn("group:create", bucket["permissions"])
        self.assertIn("system.Authenticated", bucket["permissions"]["group:create"])

        # Create Alice's friend group with Bob and Mary
        group_url = "{}/groups/alices-friends".format(bucket_url)
        resp = self.session.put(
            group_url, json={"data": {"members": [mary_user_id, bob_user_id]}}, auth=alice_auth
        )
        self.assertEqual(resp.status_code, 201)

        # Give Alice's task permission for that group
        group_id = "/buckets/{}/groups/alices-friends".format(bucket_id)
        resp = self.session.patch(
            record_url, json={"permissions": {"read": [group_id]}}, auth=alice_auth
        )
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertIn(group_id, record["permissions"]["read"])

        # Try to access Alice's task with Mary
        resp = self.session.get(record_url, auth=mary_auth)
        self.assertEqual(resp.status_code, 200)

        # Check that Mary's collection_get sees Alice's task
        resp = self.session.get(collection_url, auth=mary_auth)
        self.assertEqual(resp.status_code, 200)
        records = resp.json()["data"]
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]["id"], alice_task_id)

        # Check that Bob's collection_get sees both his and Alice's tasks
        resp = self.session.get(collection_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)
        records = resp.json()["data"]
        self.assertEqual(len(records), 2)
        records_ids = [r["id"] for r in records]
        self.assertIn(alice_task_id, records_ids)
        self.assertIn(bob_task_id, records_ids)

    def test_check_for_lists(self):
        # List buckets should not be forbidden
        resp = self.session.get("{}/buckets".format(self.server_url))
        self.assertEqual(resp.status_code, 200)



from pathlib import Path

from pyramid.httpexceptions import HTTPTemporaryRedirect
from pyramid.static import static_view

from .views import admin_home_view


def includeme(config):
    admin_assets_path = config.registry.settings["admin_assets_path"]
    if not admin_assets_path:
        # Use bundled admin.
        admin_assets_path = "kinto.plugins.admin:build"
        version_file_parent = Path(__file__).parent
    else:
        version_file_parent = Path(admin_assets_path)

    admin_version = (version_file_parent / "VERSION").read_text().strip()

    # Expose capability.
    config.add_api_capability(
        "admin",
        version=admin_version,
        description="Serves the admin console.",
        url="https://github.com/Kinto/kinto-admin/",
    )

    config.add_route("admin_home", "/admin/")
    config.add_view(admin_home_view, route_name="admin_home")

    build_dir = static_view(admin_assets_path, use_subpath=True)
    config.add_route("catchall_static", "/admin/*subpath")
    config.add_view(build_dir, route_name="catchall_static")

    # Setup redirect without trailing slash.
    def admin_redirect_view(request):
        raise HTTPTemporaryRedirect(request.path + "/")

    config.add_route("admin_redirect", "/admin")
    config.add_view(admin_redirect_view, route_name="admin_redirect")



import codecs
import logging
import os
from time import strftime

from kinto import __version__
from kinto.core import utils as core_utils


logger = logging.getLogger(__name__)

HERE = os.path.dirname(__file__)


def render_template(template, destination, **kwargs):
    template = os.path.join(HERE, template)
    folder = os.path.dirname(destination)

    if folder and not os.path.exists(folder):
        os.makedirs(folder)

    logger.info(f"Created config {os.path.abspath(destination)}")

    with codecs.open(template, "r", encoding="utf-8") as f:
        raw_template = f.read()
        rendered = raw_template.format_map(kwargs)
        with codecs.open(destination, "w+", encoding="utf-8") as output:
            output.write(rendered)


postgresql_url = "postgresql://postgres:postgres@localhost/postgres"

backend_to_values = {
    "postgresql": {
        "storage_backend": "kinto.core.storage.postgresql",
        "storage_url": postgresql_url,
        "permission_backend": "kinto.core.permission.postgresql",
        "permission_url": postgresql_url,
    },
    "memory": {
        "storage_backend": "kinto.core.storage.memory",
        "storage_url": "",
        "permission_backend": "kinto.core.permission.memory",
        "permission_url": "",
    },
}

cache_backend_to_values = {
    "postgresql": {"cache_backend": "kinto.core.cache.postgresql", "cache_url": postgresql_url},
    "memcached": {
        "cache_backend": "kinto.core.cache.memcached",
        "cache_url": "127.0.0.1:11211 127.0.0.2:11211",
    },
    "memory": {"cache_backend": "kinto.core.cache.memory", "cache_url": ""},
}


def init(config_file, backend, cache_backend, host="127.0.0.1"):
    values = {}

    values["host"] = host
    values["secret"] = core_utils.random_bytes_hex(32)
    values["bucket_id_salt"] = core_utils.random_bytes_hex(32)

    values["kinto_version"] = __version__
    values["config_file_timestamp"] = str(strftime("%a, %d %b %Y %H:%M:%S %z"))

    values.update(backend_to_values[backend])
    values.update(cache_backend_to_values[cache_backend])

    render_template("kinto.tpl", config_file, **values)



"""
kinto.core.scripts: utilities to build admin scripts for kinto-based services
"""

import logging

from pyramid.settings import asbool


logger = logging.getLogger(__name__)


def migrate(env, dry_run=False):
    """
    User-friendly frontend to run database migrations.
    """
    registry = env["registry"]
    settings = registry.settings
    readonly_backends = ("storage", "permission")
    readonly_mode = asbool(settings.get("readonly", False))

    for backend in ("cache", "storage", "permission"):
        if hasattr(registry, backend):
            if readonly_mode and backend in readonly_backends:
                message = f"Cannot migrate the {backend} backend while in readonly mode."
                logger.error(message)
            else:
                getattr(registry, backend).initialize_schema(dry_run=dry_run)


def flush_cache(env):
    registry = env["registry"]
    registry.cache.flush()
    logger.info("Cache has been cleared.")
    return 0



import random
from base64 import b64decode, b64encode
from unittest import mock
from urllib.parse import parse_qs, urlparse

from pyramid.httpexceptions import HTTPBadRequest

from kinto.core.utils import json

from . import BaseTest


class BasePaginationTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()

        indices = list(range(20))
        random.shuffle(indices)
        for i in indices:
            obj = {"title": "MoFo #{0:02}".format(i), "status": i % 4, "unread": (i % 2 == 0)}
            if i % 3 == 0:
                obj["optional"] = True
            self.model.create_object(obj)

        self.validated = self.resource.request.validated

    def _setup_next_page(self):
        next_page = self.last_response.headers["Next-Page"]
        url_fragments = urlparse(next_page)
        queryparams = parse_qs(url_fragments.query)
        self.validated["querystring"]["_token"] = queryparams["_token"][0]
        self.validated["querystring"]["_limit"] = int(queryparams["_limit"][0])
        self.last_response.headers = {}
        return queryparams


class PaginationTest(BasePaginationTest):
    def test_return_data(self):
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 20)

    def test_handle_limit(self):
        self.validated["querystring"] = {"_limit": 10}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 10)

    def test_handle_forced_limit(self):
        with mock.patch.dict(self.resource.request.registry.settings, [("paginate_by", 10)]):
            result = self.resource.plural_get()
            self.assertEqual(len(result["data"]), 10)

    def test_forced_limit_has_precedence_over_provided_limit(self):
        with mock.patch.dict(self.resource.request.registry.settings, [("paginate_by", 5)]):
            self.validated["querystring"] = {"_limit": 10}
            result = self.resource.plural_get()
            self.assertEqual(len(result["data"]), 5)

    def test_return_total_objects_in_headers(self):
        self.validated["querystring"] = {"_limit": 5}
        self.resource.plural_head()
        headers = self.last_response.headers
        count = headers["Total-Objects"]
        self.assertEqual(int(count), 20)

    def test_return_next_page_url_is_given_in_headers(self):
        self.validated["querystring"] = {"_limit": 10}
        self.resource.plural_get()
        self.assertIn("Next-Page", self.last_response.headers)

    def test_next_page_url_has_got_querystring(self):
        self.validated["querystring"] = {"_limit": 10}
        self.resource.plural_get()
        queryparams = self._setup_next_page()
        self.assertIn("_limit", queryparams)
        self.assertIn("_token", queryparams)

    def test_next_page_url_gives_distinct_objects(self):
        self.validated["querystring"] = {"_limit": 10}
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        results_id1 = set([x["id"] for x in results1["data"]])
        results_id2 = set([x["id"] for x in results2["data"]])
        self.assertFalse(results_id1.intersection(results_id2))

    def test_next_page_url_gives_distinct_objects_with_forced_limit(self):
        with mock.patch.dict(self.resource.request.registry.settings, [("paginate_by", 5)]):
            results1 = self.resource.plural_get()
            self._setup_next_page()
            results2 = self.resource.plural_get()

            results_id1 = set([x["id"] for x in results1["data"]])
            results_id2 = set([x["id"] for x in results2["data"]])
            self.assertFalse(results_id1.intersection(results_id2))

    def test_twice_the_same_next_page(self):
        self.validated["querystring"] = {"_limit": 10}
        objects = self.resource.plural_get()
        first_ids = [r["id"] for r in objects["data"]]
        objects = self.resource.plural_get()
        second_ids = [r["id"] for r in objects["data"]]
        self.assertEqual(first_ids, second_ids)

    def test_stops_giving_next_page_at_the_end_of_first_page(self):
        self.resource.plural_get()
        self.assertNotIn("Next-Page", self.last_response.headers)

    def test_stops_giving_next_page_at_the_end_sets(self):
        self.validated["querystring"] = {"_limit": 11}
        self.resource.plural_get()
        self._setup_next_page()
        self.resource.plural_get()
        self.assertNotIn("Next-Page", self.last_response.headers)

    def test_stops_giving_next_page_at_the_end_sets_on_exact_limit(self):
        self.validated["querystring"] = {"_limit": 10}
        self.resource.plural_get()
        self._setup_next_page()
        self.resource.plural_get()
        self.assertNotIn("Next-Page", self.last_response.headers)

    def test_handle_simple_sorting(self):
        self.validated["querystring"] = {"_sort": ["-status"], "_limit": 20}
        expected_results = self.resource.plural_get()
        self.validated["querystring"]["_limit"] = 10
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        self.assertEqual(expected_results["data"], results1["data"] + results2["data"])

    def test_handle_multiple_sorting(self):
        self.validated["querystring"] = {"_sort": ["-status", "title"], "_limit": 20}
        expected_results = self.resource.plural_get()
        self.validated["querystring"]["_limit"] = 10
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        self.assertEqual(expected_results["data"], results1["data"] + results2["data"])

    def test_handle_filtering_sorting(self):
        self.validated["querystring"] = {"_sort": ["-status", "title"], "status": 2, "_limit": 20}
        expected_results = self.resource.plural_get()
        self.validated["querystring"]["_limit"] = 3
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        self.assertEqual(expected_results["data"], results1["data"] + results2["data"])

    def test_handle_sorting_desc(self):
        self.validated["querystring"] = {"_sort": ["status", "-title"], "_limit": 20}
        expected_results = self.resource.plural_get()
        self.validated["querystring"]["_limit"] = 10
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        self.assertEqual(expected_results["data"], results1["data"] + results2["data"])

    def test_handle_since(self):
        self.validated["querystring"] = {"_since": 123, "_limit": 20}
        expected_results = self.resource.plural_get()
        self.validated["querystring"]["_limit"] = 10
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        self.assertEqual(expected_results["data"], results1["data"] + results2["data"])

    def test_token_wrong_base64(self):
        self.validated["querystring"] = {"_since": 123, "_limit": 20, "_token": "123"}
        self.assertRaises(HTTPBadRequest, self.resource.plural_get)

    def test_token_wrong_json(self):
        self.validated["querystring"] = {
            "_since": 123,
            "_limit": 20,
            "_token": b64encode('{"toto":'.encode("ascii")).decode("ascii"),
        }
        self.assertRaises(HTTPBadRequest, self.resource.plural_get)

    def test_token_wrong_json_fields(self):
        badtoken = '{"toto": {"tutu": 1}}'
        self.validated["querystring"] = {
            "_since": 123,
            "_limit": 20,
            "_token": b64encode(badtoken.encode("ascii")).decode("ascii"),
        }
        self.assertRaises(HTTPBadRequest, self.resource.plural_get)

    def test_raises_bad_request_if_token_has_bad_data_structure(self):
        invalid_token = json.dumps([[("last_modified", 0, ">")]])
        self.validated["querystring"] = {
            "_since": 123,
            "_limit": 20,
            "_token": b64encode(invalid_token.encode("ascii")).decode("ascii"),
        }
        self.assertRaises(HTTPBadRequest, self.resource.plural_get)

    def test_next_page_url_works_with_optional_fields(self):
        self.validated["querystring"] = {"_limit": 10, "_sort": ["-optional"]}
        results1 = self.resource.plural_get()
        self._setup_next_page()
        results2 = self.resource.plural_get()
        results_id1 = set([x["id"] for x in results1["data"]])
        results_id2 = set([x["id"] for x in results2["data"]])
        self.assertFalse(results_id1.intersection(results_id2))


class PaginatedDeleteTest(BasePaginationTest):
    def test_handle_limit_on_delete(self):
        self.validated["querystring"] = {"_limit": 3}
        result = self.resource.plural_delete()
        self.assertEqual(len(result["data"]), 3)

    def test_paginated_delete(self):
        all_objects = self.resource.plural_get()
        expected_ids = [r["id"] for r in all_objects["data"]]
        # Page 1
        self.validated["querystring"]["_limit"] = 10
        results1 = self.resource.plural_delete()
        results1_ids = [r["id"] for r in results1["data"]]
        self._setup_next_page()
        # Page 2
        results2 = self.resource.plural_delete()
        results2_ids = [r["id"] for r in results2["data"]]
        self.assertEqual(expected_ids, results1_ids + results2_ids)

    def test_paginated_delete_second_to_last_gets_next_header(self):
        self.resource.plural_head()
        get_all_headers = self.last_response.headers
        count = int(get_all_headers["Total-Objects"]) - 1

        self.validated["querystring"] = {"_limit": 1}
        headers = []
        for i in range(count):
            self.resource.plural_delete()
            headers.append(self.last_response.headers)
            self._setup_next_page()

        self.resource.plural_delete()
        headers.append(self.last_response.headers)

        self.assertIn("Next-Page", headers[count - 1])
        self.assertNotIn("Next-Page", headers[count])

    def test_token_cannot_be_reused_twice(self):
        self.resource.request.method = "DELETE"
        self.validated["querystring"]["_limit"] = 3
        self.resource.plural_delete()
        self._setup_next_page()
        self.resource.plural_delete()
        # Reuse previous token.
        self.assertRaises(HTTPBadRequest, self.resource.plural_delete)


class BuildPaginationTokenTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()
        self.obj = {
            "id": 1,
            "status": 2,
            "unread": True,
            "last_modified": 1234,
            "title": "Title",
            "nested": {"subvalue": 42},
            "nested.other": {"subvalue": 43},
        }

    def test_token_contains_current_offset(self):
        token = self.resource._build_pagination_token([("last_modified", -1)], self.obj, 42)
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(tokeninfo["offset"], 42)

    def test_no_sorting_default_to_modified_field(self):
        token = self.resource._build_pagination_token([("last_modified", -1)], self.obj, 42)
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertDictEqual(tokeninfo["last_object"], {"last_modified": 1234})

    def test_sorting_handle_both_rules(self):
        token = self.resource._build_pagination_token(
            [("status", -1), ("last_modified", -1)], self.obj, 34
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertDictEqual(tokeninfo["last_object"], {"last_modified": 1234, "status": 2})

    def test_sorting_handle_ordering_direction(self):
        token = self.resource._build_pagination_token(
            [("status", 1), ("last_modified", 1)], self.obj, 32
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(tokeninfo["last_object"], {"last_modified": 1234, "status": 2})

    def test_multiple_sorting_keep_all(self):
        token = self.resource._build_pagination_token(
            [("status", 1), ("title", -1), ("last_modified", -1)], self.obj, 31
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(
            tokeninfo["last_object"], {"last_modified": 1234, "status": 2, "title": "Title"}
        )

    def test_sorting_on_nested_field(self):
        token = self.resource._build_pagination_token(
            [("nested.subvalue", -1), ("title", 1)], self.obj, 88
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(tokeninfo["last_object"], {"title": "Title", "nested.subvalue": 42})

    def test_disambiguate_fieldname_containing_dots(self):
        token = self.resource._build_pagination_token(
            [("nested.other.subvalue", -1), ("title", 1)], self.obj, 88
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(tokeninfo["last_object"], {"title": "Title", "nested.other.subvalue": 43})

    def test_strip_malformed_sort_field(self):
        token = self.resource._build_pagination_token(
            [("non.existent", -1), ("title", 1)], self.obj, 88
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(tokeninfo["last_object"], {"title": "Title"})

    def test_can_build_while_sorting_on_missing_field(self):
        token = self.resource._build_pagination_token(
            [("unknown", 1), ("title", -1), ("last_modified", -1)], self.obj, 31
        )
        tokeninfo = json.loads(b64decode(token).decode("ascii"))
        self.assertEqual(tokeninfo["last_object"], {"last_modified": 1234, "title": "Title"})



import os
import tempfile
import unittest

from kinto.plugins.admin import views as admin_views

from ..support import BaseWebTest


admin_module = os.path.dirname(admin_views.__file__)
build_folder = os.path.join(admin_module, "build")
built_index = os.path.join(build_folder, "index.html")


class AdminViewTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Simulate admin bundle without npm
        os.makedirs(build_folder, exist_ok=True)
        with open(built_index, "w") as f:
            f.write("<html><script/></html>")

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        try:
            os.remove(built_index)
        except FileNotFoundError:
            pass

    @classmethod
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = "kinto.plugins.admin"
        return settings

    def setUp(self):
        admin_views.admin_home_view.saved = None

    def test_capability_is_exposed(self):
        self.maxDiff = None
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("admin", capabilities)
        self.assertIn("version", capabilities["admin"])
        del capabilities["admin"]["version"]
        expected = {
            "description": "Serves the admin console.",
            "url": ("https://github.com/Kinto/kinto-admin/"),
        }
        self.assertEqual(expected, capabilities["admin"])

    def test_permission_endpoint_is_enabled_with_admin(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        assert "permissions_endpoint" in capabilities

    def test_admin_index_can_be_reached(self):
        self.maxDiff = None
        resp = self.app.get("/admin/")
        assert "<html>" in resp.body.decode("utf-8")

    def test_admin_redirect_without_trailing_slash(self):
        resp = self.app.get("/admin", status=307)
        self.assertTrue(resp.headers["location"].endswith("/admin/"))

    def test_admin_has_csp_header(self):
        resp = self.app.get("/admin/")
        assert "default-src 'self'" in resp.headers["Content-Security-Policy"]
        # The cached version too.
        resp = self.app.get("/admin/")
        assert "default-src 'self'" in resp.headers["Content-Security-Policy"]


class OverriddenAdminViewTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def make_app(cls, *args, **kwargs):
        cls.tmp_dir = tempfile.TemporaryDirectory()
        with open(os.path.join(cls.tmp_dir.name, "VERSION"), "w") as f:
            f.write("42.0.0")
        with open(os.path.join(cls.tmp_dir.name, "index.html"), "w") as f:
            f.write("mine!")
        with open(os.path.join(cls.tmp_dir.name, "script.js"), "w") as f:
            f.write("kiddy")
        return super().make_app(*args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.tmp_dir.cleanup()

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = "kinto.plugins.admin"
        settings["admin_assets_path"] = cls.tmp_dir.name
        return settings

    def test_admin_capability_reads_version_from_configured_folder(self):
        resp = self.app.get("/")
        self.assertEqual(resp.json["capabilities"]["admin"]["version"], "42.0.0")

    def test_admin_ui_is_served_from_configured_folder(self):
        resp = self.app.get("/admin/")
        self.assertIn("mine!", resp.body.decode("utf-8"))

    def test_assets_are_served_from_configured_folder(self):
        resp = self.app.get("/admin/script.js")
        self.assertIn("kiddy", resp.body.decode("utf-8"))

    def test_original_assets_are_not_available(self):
        self.app.get("/admin/favicon.png", status=404)



import colander
from cornice.service import get_services
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core import Service
from kinto.core.openapi import OpenAPI


openapi = Service(name="openapi", path="/__api__", description="OpenAPI description")


class OpenAPIResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown="preserve"))


openapi_response_schemas = {
    "200": OpenAPIResponseSchema(
        description="Return an OpenAPI description of the running instance."
    )
}


@openapi.get(
    permission=NO_PERMISSION_REQUIRED,
    response_schemas=openapi_response_schemas,
    tags=["Utilities"],
    operation_id="get_openapi_spec",
)
def openapi_view(request):
    # Only build json once
    try:
        return openapi_view.__json__
    except AttributeError:
        openapi_view.__json__ = OpenAPI(get_services(), request).generate()
        return openapi_view.__json__



import warnings
from unittest import mock

import webtest
from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.exceptions import ConfigurationError

import kinto.core
from kinto.core import initialization
from kinto.core.testing import unittest


class InitializationTest(unittest.TestCase):
    def test_fails_if_no_version_is_specified(self):
        config = Configurator()
        self.assertRaises(ConfigurationError, kinto.core.initialize, config)

    def test_uses_the_version_for_prefix(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.route_prefix, "v0")

    def test_set_the_project_version_if_specified(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.registry.settings["project_version"], "0.0.1")

    def test_project_version_uses_setting_if_specified(self):
        config = Configurator(settings={"name.project_version": "1.0.0"})
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.registry.settings["project_version"], "1.0.0")

    def test_http_api_version_relies_on_project_version_by_default(self):
        config = Configurator()
        kinto.core.initialize(config, "0.1.0", "name")
        self.assertEqual(config.registry.settings["http_api_version"], "0.1")

    def test_http_api_version_uses_setting_if_specified(self):
        config = Configurator(settings={"name.http_api_version": "1.3"})
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.registry.settings["http_api_version"], "1.3")

    def test_warns_if_settings_prefix_is_empty(self):
        config = Configurator(settings={"kinto.settings_prefix": ""})
        with mock.patch("kinto.core.initialization.warnings.warn") as mocked:
            kinto.core.initialize(config, "0.0.1")
            error_msg = "No value specified for `settings_prefix`"
            mocked.assert_any_call(error_msg)

    def test_warns_if_settings_prefix_is_missing(self):
        config = Configurator()
        with mock.patch("kinto.core.initialization.warnings.warn") as mocked:
            kinto.core.initialize(config, "0.0.1")
            error_msg = "No value specified for `settings_prefix`"
            mocked.assert_any_call(error_msg)

    def test_set_the_settings_prefix_if_specified(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "kinto")
        self.assertEqual(config.registry.settings["settings_prefix"], "kinto")

    def test_set_the_settings_prefix_from_settings_even_if_specified(self):
        config = Configurator(settings={"kinto.settings_prefix": "kinto"})
        kinto.core.initialize(config, "0.0.1", "readinglist")
        self.assertEqual(config.registry.settings["settings_prefix"], "kinto")

    def test_warns_if_not_https(self):
        error_msg = "HTTPS is not enabled"
        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")

            config = Configurator(
                settings={"settings_prefix": "kinto", "kinto.http_scheme": "https"}
            )
            kinto.core.initialize(config, "0.0.1")
            self.assertFalse(any(error_msg in str(w.message) for w in warns))

            config = Configurator(settings={"kinto.http_scheme": "http"})
            kinto.core.initialize(config, "0.0.1")
            self.assertTrue(any(error_msg in str(w.message) for w in warns))

    def test_default_settings_are_overriden_if_specified_in_initialize(self):
        config = Configurator()
        defaults = {"paginate_by": 102}
        kinto.core.initialize(config, "0.0.1", "prefix", default_settings=defaults)
        self.assertEqual(config.registry.settings["paginate_by"], 102)

    def test_default_settings_are_overriden_by_application(self):
        config = Configurator(settings={"prefix.paginate_by": 10})
        kinto.core.initialize(config, "0.0.1", "prefix")
        self.assertEqual(config.registry.settings["paginate_by"], 10)

    def test_specified_default_settings_are_overriden_by_application(self):
        config = Configurator(settings={"settings_prefix.paginate_by": 5})
        defaults = {"settings_prefix.paginate_by": 10}
        kinto.core.initialize(config, "0.0.1", "settings_prefix", default_settings=defaults)
        self.assertEqual(config.registry.settings["paginate_by"], 5)

    def test_backends_are_not_instantiated_by_default(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        kinto.core.initialize(config, "0.0.1", "settings_prefix")
        self.assertFalse(hasattr(config.registry, "storage"))
        self.assertFalse(hasattr(config.registry, "cache"))
        self.assertFalse(hasattr(config.registry, "permission"))

    def test_backends_type_is_checked_when_instantiated(self):
        def config_fails(settings):
            config = Configurator(settings=settings)
            with self.assertRaises(ConfigurationError):
                kinto.core.initialize(config, "0.0.1", "settings_prefix")

        config_fails({"settings_prefix.storage_backend": "kinto.core.cache.memory"})
        config_fails({"settings_prefix.cache_backend": "kinto.core.storage.memory"})
        config_fails({"settings_prefix.permission_backend": "kinto.core.storage.memory"})

    def test_environment_values_override_configuration(self):
        import os

        envkey = "KINTO_SETTINGS_PREFIX"
        os.environ[envkey] = "abc"

        config = Configurator(settings={"kinto.settings_prefix": "kinto"})
        kinto.core.initialize(config, "0.0.1")

        os.environ.pop(envkey)

        prefix_used = config.registry.settings["settings_prefix"]
        self.assertEqual(prefix_used, "abc")


class ProjectSettingsTest(unittest.TestCase):
    def settings(self, provided):
        config = Configurator(settings=provided)
        kinto.core.initialize(config, "1.0", "myproject")
        return config.get_settings()

    def test_uses_unprefixed_name(self):
        settings = {"paginate_by": 3.14}
        self.assertEqual(self.settings(settings)["paginate_by"], 3.14)

    def test_uses_settings_prefix(self):
        settings = {"myproject.paginate_by": 42}
        self.assertEqual(self.settings(settings)["paginate_by"], 42)

    def test_does_raise_valueerror_if_multiple_entries_are_equal(self):
        settings = {"paginate_by": 42, "myproject.paginate_by": 42}
        self.settings(settings)  # Not raising.

    def test_does_raise_valueerror_if_entries_are_not_hashable(self):
        settings = {"events.listeners": ["Rémy", 42]}
        self.settings(settings)  # Not raising.

    def test_raises_valueerror_if_different_multiple_entries(self):
        settings = {"paginate_by": 42, "myproject.paginate_by": 3.14}
        with self.assertRaises(ValueError):
            self.settings(settings)

    def test_environment_can_specify_settings_prefix(self):
        import os

        envkey = "MYPROJECT_CACHE_BACKEND"  # MYPROJECT_ prefix
        os.environ[envkey] = "kinto.core.cache.memory"
        settings = {"kinto.cache_backend": "kinto.core.cache.memcached"}
        value = self.settings(settings)["cache_backend"]
        os.environ.pop(envkey)
        self.assertEqual(value, "kinto.core.cache.memory")


class ApplicationWrapperTest(unittest.TestCase):
    @unittest.skipIf(initialization.newrelic is None, "newrelic is not installed.")
    @mock.patch("kinto.core.initialization.newrelic.agent")
    def test_newrelic_is_included_if_defined(self, mocked_newrelic):
        settings = {"newrelic_config": "/foo/bar.ini", "newrelic_env": "test"}
        mocked_newrelic.WSGIApplicationWrapper.return_value = "wrappedApp"
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_newrelic.initialize.assert_called_with("/foo/bar.ini", "test")
        self.assertEqual(app, "wrappedApp")

    @unittest.skipIf(initialization.newrelic is None, "newrelic is not installed.")
    @mock.patch("kinto.core.initialization.newrelic.agent")
    def test_newrelic_is_not_included_if_set_to_false(self, mocked_newrelic):
        settings = {"newrelic_config": False}
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_newrelic.initialize.assert_not_called()
        self.assertEqual(app, mock.sentinel.app)

    @mock.patch("kinto.core.initialization.ProfilerMiddleware")
    def test_profiler_is_not_installed_if_set_to_false(self, mocked_profiler):
        settings = {"profiler_enabled": False}
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_profiler.initialize.assert_not_called()
        self.assertEqual(app, mock.sentinel.app)

    @mock.patch("kinto.core.initialization.ProfilerMiddleware")
    def test_profiler_is_installed_if_set_to_true(self, mocked_profiler):
        settings = {"profiler_enabled": True, "profiler_dir": "/tmp/path"}
        mocked_profiler.return_value = "wrappedApp"
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)

        mocked_profiler.assert_called_with(
            mock.sentinel.app, restrictions="*kinto.core*", profile_dir="/tmp/path"
        )

        self.assertEqual(app, "wrappedApp")

    def test_load_default_settings_handle_prefix_attributes(self):
        config = mock.MagicMock()

        settings = {"settings_prefix": "myapp"}

        config.get_settings.return_value = settings
        initialization.load_default_settings(
            config,
            {
                "multiauth.policies": "basicauth",
                "myapp.http_scheme": "https",
                "myapp.http_host": "localhost:8888",
            },
        )

        self.assertIn("multiauth.policies", settings)
        self.assertNotIn("policies", settings)
        self.assertEqual(settings["multiauth.policies"], "basicauth")

        self.assertIn("http_scheme", settings)
        self.assertNotIn("kinto.http_scheme", settings)
        self.assertEqual(settings["http_scheme"], "https")

        self.assertIn("http_host", settings)
        self.assertNotIn("myapp.http_host", settings)
        self.assertEqual(settings["http_host"], "localhost:8888")

    def test_load_default_settings_converts_to_native_correctly(self):
        config = mock.MagicMock()

        settings = {"settings_prefix": "myapp"}

        config.get_settings.return_value = settings
        initialization.load_default_settings(config, {"myapp.my_cool_setting": '"1.2"'})

        self.assertEqual(settings["my_cool_setting"], "1.2")


class SentryTest(unittest.TestCase):
    @unittest.skipIf(initialization.sentry_sdk is None, "sentry is not installed.")
    def test_sentry_not_enabled_by_default(self):
        config = Configurator()
        with mock.patch("sentry_sdk.init") as mocked:
            kinto.core.initialize(config, "0.0.1")

        self.assertFalse(mocked.called)

    @unittest.skipIf(initialization.sentry_sdk is None, "sentry is not installed.")
    def test_sentry_enabled_if_sentry_dsn_is_set(self):
        config = Configurator(
            settings={
                "sentry_dsn": "https://notempty",
                "sentry_env": "local",
            }
        )
        with mock.patch("sentry_sdk.init") as mocked:
            kinto.core.initialize(config, "0.0.1")

        init_call = mocked.call_args_list[0]
        self.assertEqual(init_call[0][0], "https://notempty")
        self.assertEqual(init_call[1]["environment"], "local")

    @unittest.skipIf(initialization.sentry_sdk is None, "sentry is not installed.")
    def test_message_is_sent_on_startup(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings(
            {
                "sentry_dsn": "https://notempty",
            }
        )

        with mock.patch("sentry_sdk.init"):
            kinto.core.initialize(config, "0.0.1", "name")

            with mock.patch("sentry_sdk.capture_message") as mocked:
                config.make_wsgi_app()

        mocked.assert_called_with("Running  0.0.1.", "info")

    @unittest.skipIf(initialization.sentry_sdk is None, "sentry is not installed.")
    def unexpected_exceptions_are_reported(self):
        with mock.patch("kinto.core.views.hello.get_eos", side_effect=ValueError):
            with mock.patch("sentry_sdk.hub.Hub.capture_event") as mocked:
                resp = self.app.get("/")
        assert resp.status == 500
        assert len(mocked.call_args_list) > 0


class StatsDConfigurationTest(unittest.TestCase):
    def setUp(self):
        settings = {
            **kinto.core.DEFAULT_SETTINGS,
            "statsd_url": "udp://host:8080",
            "multiauth.policies": "basicauth",
        }
        self.config = Configurator(settings=settings)
        self.config.registry.storage = {}
        self.config.registry.cache = {}
        self.config.registry.permission = {}

        patch = mock.patch("kinto.core.statsd.load_from_config")
        self.mocked = patch.start()
        self.addCleanup(patch.stop)

    def test_statsd_isnt_called_if_statsd_url_is_not_set(self):
        self.config.add_settings({"statsd_url": None})
        initialization.setup_statsd(self.config)
        self.mocked.assert_not_called()

    def test_statsd_is_set_to_none_if_statsd_url_not_set(self):
        self.config.add_settings({"statsd_url": None})
        initialization.setup_statsd(self.config)
        self.assertEqual(self.config.registry.statsd, None)

    def test_statsd_is_called_if_statsd_url_is_set(self):
        initialization.setup_statsd(self.config)
        self.mocked.assert_called_with(self.config)
        # See `tests/core/test_statsd.py` for instantiation tests.

    def test_statsd_is_expose_in_the_registry_if_url_is_set(self):
        initialization.setup_statsd(self.config)
        self.assertEqual(self.config.registry.statsd, self.mocked.return_value)

    def test_statsd_is_set_on_cache(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix="backend")

    def test_statsd_is_set_on_storage(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix="backend")

    def test_statsd_is_set_on_permission(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix="backend")

    def test_statsd_is_set_on_authentication(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call(None, prefix="authentication")

    def test_statsd_counts_nothing_on_anonymous_requests(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get("/")
        self.assertFalse(self.mocked.count.called)

    def test_statsd_counts_views_and_methods(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get("/v0/__heartbeat__")
        self.mocked().count.assert_any_call("view.heartbeat.GET")

    def test_statsd_counts_unknown_urls(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get("/v0/coucou", status=404)
        self.assertFalse(self.mocked.count.called)

    @mock.patch("kinto.core.utils.hmac_digest")
    def test_statsd_counts_unique_users(self, digest_mocked):
        digest_mocked.return_value = "mat"
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        headers = {"Authorization": "Basic bWF0Og=="}
        app.get("/v0/", headers=headers)
        self.mocked().count.assert_any_call("users", unique="basicauth.mat")

    def test_statsd_counts_authentication_types(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        headers = {"Authorization": "Basic bWF0Og=="}
        app.get("/v0/", headers=headers)
        self.mocked().count.assert_any_call("authn_type.basicauth")


class RequestsConfigurationTest(unittest.TestCase):
    def _get_app(self, settings={}):
        app_settings = {
            "storage_backend": "kinto.core.storage.memory",
            "cache_backend": "kinto.core.cache.memory",
        }
        app_settings.update(**settings)
        config = Configurator(settings=app_settings)
        kinto.core.initialize(config, "0.0.1", "name")
        return webtest.TestApp(config.make_wsgi_app())

    def test_requests_have_a_bound_data_attribute(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")

        def on_new_request(event):
            data = event.request.bound_data
            self.assertEqual(data, {})
            self.assertEqual(id(data), id(event.request.bound_data))

        config.add_subscriber(on_new_request, NewRequest)
        app = webtest.TestApp(config.make_wsgi_app())
        app.get("/v0/")

    def test_subrequests_share_parent_bound_data(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")

        bound_datas = set()

        def on_new_request(event):
            bound_datas.add(id(event.request.bound_data))

        config.add_subscriber(on_new_request, NewRequest)
        app = webtest.TestApp(config.make_wsgi_app())
        app.post_json("/v0/batch", {"requests": [{"path": "/"}]})
        self.assertEqual(len(bound_datas), 1)

    def test_by_default_relies_on_pyramid_application_url(self):
        app = self._get_app()
        resp = app.get("/v0/")
        self.assertEqual(resp.json["url"], "http://localhost/v0/")

    def test_by_default_relies_on_incoming_headers(self):
        app = self._get_app()
        resp = app.get("/v0/", headers={"Host": "server:8888"})
        self.assertEqual(resp.json["url"], "http://server:8888/v0/")

    def test_by_default_relies_on_wsgi_environment(self):
        app = self._get_app()
        environ = {"wsgi.url_scheme": "https", "HTTP_HOST": "server:44311"}
        resp = app.get("/v0/", extra_environ=environ)
        self.assertEqual(resp.json["url"], "https://server:44311/v0/")

    def test_http_scheme_overrides_the_wsgi_environment(self):
        app = self._get_app({"http_scheme": "http2"})
        environ = {"wsgi.url_scheme": "https"}
        resp = app.get("/v0/", extra_environ=environ)
        self.assertEqual(resp.json["url"], "http2://localhost:80/v0/")

    def test_http_host_overrides_the_wsgi_environment(self):
        app = self._get_app({"http_host": "server"})
        environ = {"HTTP_HOST": "elb:44311"}
        resp = app.get("/v0/", extra_environ=environ)
        self.assertEqual(resp.json["url"], "http://server/v0/")

    def test_http_host_overrides_the_request_headers(self):
        app = self._get_app({"http_host": "server"})
        resp = app.get("/v0/", headers={"Host": "elb:8888"})
        self.assertEqual(resp.json["url"], "http://server/v0/")


class PluginsTest(unittest.TestCase):
    def test_kinto_core_includes_are_included_manually(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings({"includes": "elastic history"})
        config.route_prefix = "v2"

        with mock.patch.object(config, "include"):
            with mock.patch.object(config, "scan"):
                kinto.core.includeme(config)

                config.include.assert_any_call("elastic")
                config.include.assert_any_call("history")

    def make_app(self, settings=None):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings(
            {
                "permission_backend": "kinto.core.permission.memory",
                "includes": "tests.core.testplugin",
                "multiauth.policies": "basicauth",
            },
            **(settings or {}),
        )
        kinto.core.initialize(config, "0.0.1", "name")
        return webtest.TestApp(config.make_wsgi_app())

    def test_plugin_can_define_protected_views(self):
        app = self.make_app()
        app.post("/v0/attachment", status=401)
        headers = {"Authorization": "Basic bWF0OjE="}
        app.post("/v0/attachment", headers=headers, status=403)

    def test_plugin_benefits_from_cors_setup(self):
        app = self.make_app()
        headers = {"Origin": "lolnet.org", "Access-Control-Request-Method": "POST"}
        resp = app.options("/v0/attachment", headers=headers, status=200)
        self.assertIn("Access-Control-Allow-Origin", resp.headers)

    def test_write_http_methods_are_rejected_if_readonly(self):
        app = self.make_app({"readonly": True})
        # `pytest.mark.parametrize` is not available on `TestCase` methods.
        for method, status in [
            ("GET", 200),
            ("HEAD", 200),
            ("OPTIONS", 200),
            ("POST", 405),
            ("PUT", 405),
            ("PATCH", 405),
            ("DELETE", 405),
        ]:
            app.request(
                "/v0/log",
                method=method,
                status=status,
                headers={
                    "Access-Control-Request-Method": method,
                    "Origin": "http://server.com",
                },
            )



import unittest

from kinto.core.testing import get_user_headers

from .support import MINIMALIST_BUCKET, MINIMALIST_COLLECTION, MINIMALIST_RECORD, BaseWebTest


class CollectionViewTest(BaseWebTest, unittest.TestCase):
    collections_url = "/buckets/beers/collections"
    collection_url = "/buckets/beers/collections/barley"

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        resp = self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        self.record = resp.json["data"]

    def test_collection_endpoint_lists_them_all(self):
        resp = self.app.get(self.collections_url, headers=self.headers)
        records = resp.json["data"]
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]["id"], "barley")

    def test_collections_can_be_put_with_simple_name(self):
        self.assertEqual(self.record["id"], "barley")

    def test_collections_name_should_be_simple(self):
        self.app.put_json(
            "/buckets/beers/collections/__barley__",
            MINIMALIST_COLLECTION,
            headers=self.headers,
            status=400,
        )

    def test_collections_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, "Content-Type": "text/plain"}
        self.app.put(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=headers, status=415
        )

    def test_unknown_bucket_raises_403(self):
        other_bucket = self.collections_url.replace("beers", "sodas")
        self.app.get(other_bucket, headers=self.headers, status=403)

    def test_unknown_collection_raises_404(self):
        other_collection = self.collection_url.replace("barley", "pills")
        resp = self.app.get(other_collection, headers=self.headers, status=404)
        self.assertEqual(resp.json["details"]["id"], "pills")
        self.assertEqual(resp.json["details"]["resource_name"], "collection")

    def test_collections_are_isolated_by_bucket(self):
        other_bucket = self.collection_url.replace("beers", "sodas")
        self.app.put_json("/buckets/sodas", MINIMALIST_BUCKET, headers=self.headers)
        self.app.get(other_bucket, headers=self.headers, status=404)

    def test_create_permissions_can_be_added_on_collections(self):
        collection = {**MINIMALIST_COLLECTION, "permissions": {"record:create": ["fxa:user"]}}
        resp = self.app.put_json(
            "/buckets/beers/collections/barley", collection, headers=self.headers, status=200
        )
        permissions = resp.json["permissions"]
        self.assertIn("fxa:user", permissions["record:create"])

    def test_wrong_create_permissions_cannot_be_added_on_collections(self):
        collection = {**MINIMALIST_COLLECTION, "permissions": {"collection:create": ["fxa:user"]}}
        self.app.put_json(
            "/buckets/beers/collections/barley", collection, headers=self.headers, status=400
        )

    def test_collections_can_handle_arbitrary_attributes(self):
        fingerprint = "5866f245a00bb3a39100d31b2f14d453"
        collection = {**MINIMALIST_COLLECTION, "data": {"fingerprint": fingerprint}}
        resp = self.app.put_json(
            "/buckets/beers/collections/barley", collection, headers=self.headers, status=200
        )
        data = resp.json["data"]
        self.assertIn("fingerprint", data)
        self.assertEqual(data["fingerprint"], fingerprint)

    def test_collections_can_be_filtered_by_arbitrary_attribute(self):
        collection = {**MINIMALIST_COLLECTION, "data": {"size": 3}}
        self.app.put_json("/buckets/beers/collections/moderator", collection, headers=self.headers)
        resp = self.app.get(
            "/buckets/beers/collections?has_size=true&min_size=2", headers=self.headers
        )
        data = resp.json["data"]
        self.assertEqual(len(data), 1)


class CollectionDeletionTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/collections/barley"

    def setUp(self):
        super().setUp()
        bucket = {
            **MINIMALIST_BUCKET,
            "permissions": {"collection:create": ["system.Everyone"], "read": ["system.Everyone"]},
        }
        self.app.put_json("/buckets/beers", bucket, headers=self.headers)
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        resp = self.app.post_json(
            self.collection_url + "/records", MINIMALIST_RECORD, headers=self.headers
        )
        self.previous_ts = resp.headers["ETag"]
        record_id = resp.json["data"]["id"]
        self.record_url = self.collection_url + "/records/{}".format(record_id)
        self.app.delete(self.collection_url, headers=self.headers)

    def test_collections_can_be_deleted(self):
        self.app.get(self.collection_url, headers=self.headers, status=404)

    def test_collections_can_be_deleted_in_bulk(self):
        alice_headers = get_user_headers("alice")
        self.app.put_json(
            "/buckets/beers/collections/1", MINIMALIST_COLLECTION, headers=self.headers
        )
        self.app.put_json(
            "/buckets/beers/collections/2", MINIMALIST_COLLECTION, headers=alice_headers
        )
        self.app.put_json(
            "/buckets/beers/collections/3", MINIMALIST_COLLECTION, headers=alice_headers
        )
        self.app.delete("/buckets/beers/collections", headers=alice_headers)
        resp = self.app.get("/buckets/beers/collections", headers=self.headers)
        self.assertEqual(len(resp.json["data"]), 1)

    def test_records_of_collection_are_deleted_too(self):
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        self.app.get(self.record_url, headers=self.headers, status=404)

        # Verify tombstones
        resp = self.app.get(
            "{}/records?_since=0".format(self.collection_url), headers=self.headers
        )
        self.assertEqual(len(resp.json["data"]), 0)

    def test_timestamps_are_refreshed_when_collection_is_recreated(self):
        # Kinto/kinto#1223
        # Recreate with same name.
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        resp = self.app.get(self.collection_url + "/records", headers=self.headers)
        records_ts = resp.headers["ETag"]
        self.assertNotEqual(self.previous_ts, records_ts)

    def test_can_be_created_after_deletion_with_if_none_match_star(self):
        headers = {**self.headers, "If-None-Match": "*"}
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=headers, status=201)

    def test_collection_permissions_are_removed_after_collection_deleted(self):
        self.assertDictEqual(self.permission.get_object_permissions(self.collection_url), {})

    def test_records_permissions_are_removed_after_collection_deleted(self):
        self.assertDictEqual(self.permission.get_object_permissions(self.record_url), {})


class CollectionCreationTest(BaseWebTest, unittest.TestCase):
    collections_url = "/buckets/beers/collections"

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)

    def test_collection_can_be_created_with_post(self):
        r = self.app.post_json(self.collections_url, MINIMALIST_COLLECTION, headers=self.headers)
        self.assertEqual(r.status_code, 201)
        self.assertTrue(len(r.json["data"]["id"]) == 8)

    def test_collection_can_be_specified_in_post(self):
        collection = "barley"
        r = self.app.post_json(
            self.collections_url, {"data": {"id": collection}}, headers=self.headers
        )
        self.assertEqual(r.status_code, 201)
        self.assertEqual(r.json["data"]["id"], collection)

    def test_collection_already_exists_post(self):
        collection = "barley"
        self.app.post_json(
            self.collections_url, {"data": {"id": collection}}, headers=self.headers
        )
        r = self.app.post_json(
            self.collections_url, {"data": {"id": collection}}, headers=self.headers
        )
        self.assertEqual(r.json["data"]["id"], collection)
        self.assertEqual(r.status_code, 200)



import os
import uuid
from unittest import mock

from pyramid import testing

from kinto.core import initialization
from kinto.core.events import ACTIONS, ResourceChanged, ResourceRead
from kinto.core.listeners import ListenerBase
from kinto.core.testing import unittest


UID = str(uuid.uuid4())


class ViewSet:
    def get_name(*args, **kw):
        return "mushrooms"


class Service:
    viewset = ViewSet()


class Match:
    cornice_services = {"watev": Service()}
    pattern = "watev"


class Request:
    path = "/1/school/students/"
    prefixed_userid = "tarek"
    matchdict = {"id": UID}
    registry = matched_route = Match()
    current_resource_name = "school"


class ListenerSetupTest(unittest.TestCase):
    def setUp(self):
        demo_patch = mock.patch("tests.core.listeners.load_from_config")
        self.addCleanup(demo_patch.stop)
        self.demo_mocked = demo_patch.start()

    def make_app(self, extra_settings={}):
        settings = {"event_listeners": "tests.core.listeners"}
        settings.update(**extra_settings)
        config = testing.setUp(settings=settings)
        config.commit()
        initialization.setup_listeners(config)
        return config

    def test_listener_module_is_specified_via_settings(self):
        self.make_app(
            {"event_listeners": "demo", "event_listeners.demo.use": "tests.core.listeners"}
        )
        self.assertTrue(self.demo_mocked.called)

    def test_listener_module_can_be_specified_via_listeners_list(self):
        self.make_app(
            {"event_listeners": "demo", "event_listeners.demo.use": "tests.core.listeners"}
        )
        self.assertTrue(self.demo_mocked.called)

    def test_callback_called_when_action_is_not_filtered(self):
        config = self.make_app(
            {"event_listeners": "demo", "event_listeners.demo.use": "tests.core.listeners"}
        )
        ev = ResourceChanged({"action": ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(ev)

        self.assertTrue(self.demo_mocked.return_value.called)

    def test_callback_is_not_called_when_action_is_filtered(self):
        config = self.make_app(
            {
                "event_listeners": "demo",
                "event_listeners.demo.use": "tests.core.listeners",
                "event_listeners.demo.actions": "delete",
            }
        )
        ev = ResourceChanged({"action": ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(ev)

        self.assertFalse(self.demo_mocked.return_value.called)

    def test_callback_called_when_resource_is_not_filtered(self):
        config = self.make_app(
            {"event_listeners": "demo", "event_listeners.demo.use": "tests.core.listeners"}
        )
        event = ResourceChanged(
            {"action": ACTIONS.CREATE.value, "resource_name": "mushroom"}, [], Request()
        )
        config.registry.notify(event)

        self.assertTrue(self.demo_mocked.return_value.called)

    def test_callback_is_not_called_when_resource_is_filtered(self):
        config = self.make_app(
            {
                "event_listeners": "demo",
                "event_listeners.demo.use": "tests.core.listeners",
                "event_listeners.demo.resources": "toad",
            }
        )
        event = ResourceChanged(
            {"action": ACTIONS.CREATE.value, "resource_name": "mushroom"}, [], Request()
        )
        config.registry.notify(event)

        self.assertFalse(self.demo_mocked.return_value.called)

    def test_callback_is_not_called_on_read_by_default(self):
        config = self.make_app(
            {"event_listeners": "demo", "event_listeners.demo.use": "tests.core.listeners"}
        )
        event = ResourceRead({"action": ACTIONS.READ.value}, [], Request())
        config.registry.notify(event)

        self.assertFalse(self.demo_mocked.return_value.called)

    def test_callback_is_called_on_read_if_specified(self):
        config = self.make_app(
            {
                "event_listeners": "demo",
                "event_listeners.demo.use": "tests.core.listeners",
                "event_listeners.demo.actions": "read",
            }
        )
        event = ResourceRead({"action": ACTIONS.READ.value}, [], Request())
        config.registry.notify(event)

        self.assertTrue(self.demo_mocked.return_value.called)

    def test_same_callback_is_called_for_read_and_write_specified(self):
        config = self.make_app(
            {
                "event_listeners": "demo",
                "event_listeners.demo.use": "tests.core.listeners",
                "event_listeners.demo.actions": "read create delete",
            }
        )
        ev = ResourceRead({"action": ACTIONS.READ.value}, [], Request())
        config.registry.notify(ev)
        ev = ResourceChanged({"action": ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(ev)

        self.assertEqual(self.demo_mocked.return_value.call_count, 2)

    def test_loading_can_read_configuration_from_environment(self):
        environ = {
            "KINTO_EVENT_LISTENERS": "kvstore",
            "KINTO_EVENT_LISTENERS_KVSTORE_USE": "tests.core.listeners",
            "KINTO_EVENT_LISTENERS_KVSTORE_URL": "demo://demo:6379/0",
            "KINTO_EVENT_LISTENERS_KVSTORE_POOL_SIZE": "5",
            "KINTO_EVENT_LISTENERS_KVSTORE_LISTNAME": "queue",
            "KINTO_EVENT_LISTENERS_KVSTORE_ACTIONS": "delete",
            "KINTO_EVENT_LISTENERS_KVSTORE_RESOURCES": "toad",
        }
        os.environ.update(**environ)

        config = self.make_app(
            {
                # With real/full initialization, these should not be necessary:
                "settings_prefix": "kinto",
                "event_listeners": "kvstore",
            }
        )

        # Listener is instantiated.
        self.assertTrue(self.demo_mocked.called)

        # Action filtering is read from ENV.
        event = ResourceChanged(
            {"action": ACTIONS.DELETE.value, "resource_name": "toad"}, [], Request()
        )
        config.registry.notify(event)
        self.assertTrue(self.demo_mocked.return_value.called)

        self.demo_mocked.reset_mock()

        # Action filtering is read from ENV.
        event = ResourceChanged({"action": ACTIONS.CREATE.value}, [], Request())
        config.registry.notify(event)
        self.assertFalse(self.demo_mocked.return_value.called)

        # Resource filtering is read from ENV.
        event = ResourceChanged(
            {"action": ACTIONS.CREATE.value, "resource_name": "mushroom"}, [], Request()
        )
        config.registry.notify(event)
        self.assertFalse(self.demo_mocked.return_value.called)

        # Clean-up.
        for k in environ.keys():
            os.environ.pop(k)


class ListenerBaseTest(unittest.TestCase):
    def test_not_implemented(self):
        # make sure we can't use the base listener
        listener = ListenerBase()
        self.assertRaises(NotImplementedError, listener, object())



import json
import os
import tempfile
import unittest
from unittest import mock

from kinto.core.views.version import version_view

from .support import BaseWebTest


class VersionViewTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        try:
            delattr(version_view, "__json__")  # Clean-up memoization.
        except AttributeError:
            pass

    tearDown = setUp

    def test_return_the_version_file_in_current_folder_if_present(self):
        content = {"version": "0.8.1"}
        fake_file = mock.mock_open(read_data=json.dumps(content))
        with mock.patch("os.path.exists"):
            with mock.patch("kinto.core.views.version.open", fake_file, create=True):
                response = self.app.get("/__version__")
                assert response.json == content

    def test_return_a_500_if_version_file_if_not_present(self):
        self.app.get("/__version__", status=500)

    def test_return_the_version_file_specified_in_setting_if_present(self):
        custom_path = tempfile.mktemp()
        content = {"foo": "lala"}
        with open(custom_path, "w") as f:
            json.dump(content, f)
        self.addCleanup(lambda: os.remove(custom_path))

        with mock.patch.dict(self.app.app.registry.settings, [("version_json_path", custom_path)]):
            response = self.app.get("/__version__")

        assert response.json == content



import re
import sys

from pyramid.exceptions import ConfigurationError

from kinto.authorization import PERMISSIONS_INHERITANCE_TREE

from .authentication import AccountsAuthenticationPolicy as AccountsPolicy
from .utils import (
    ACCOUNT_CACHE_KEY,
    ACCOUNT_POLICY_NAME,
    ACCOUNT_RESET_PASSWORD_CACHE_KEY,
    ACCOUNT_VALIDATION_CACHE_KEY,
)


__all__ = [
    "ACCOUNT_CACHE_KEY",
    "ACCOUNT_POLICY_NAME",
    "ACCOUNT_RESET_PASSWORD_CACHE_KEY",
    "ACCOUNT_VALIDATION_CACHE_KEY",
    "AccountsPolicy",
]

DOCS_URL = "https://kinto.readthedocs.io/en/stable/api/1.x/accounts.html"


def includeme(config):
    settings = config.get_settings()
    validation_enabled = settings.get("account_validation", False)
    config.add_api_capability(
        "accounts",
        description="Manage user accounts.",
        url="https://kinto.readthedocs.io/en/latest/api/1.x/accounts.html",
        validation_enabled=validation_enabled,
    )
    kwargs = {}
    if not validation_enabled:
        kwargs["ignore"] = "kinto.plugins.accounts.views.validation"
    config.scan("kinto.plugins.accounts.views", **kwargs)

    PERMISSIONS_INHERITANCE_TREE["root"].update({"account:create": {}})
    PERMISSIONS_INHERITANCE_TREE["account"] = {
        "write": {"account": ["write"]},
        "read": {"account": ["write", "read"]},
    }

    if validation_enabled:
        # Valid mailers other than the default are `debug` and `testing`
        # according to
        # https://docs.pylonsproject.org/projects/pyramid_mailer/en/latest/#debugging
        mailer = settings.get("mail.mailer", "")
        config.include("pyramid_mailer" + (f".{mailer}" if mailer else ""))

    # Check that the account policy is mentioned in config if included.
    accountClass = "AccountsPolicy"
    policy = None
    for k, v in settings.items():
        m = re.match("multiauth\\.policy\\.(.*)\\.use", k)
        if m:
            if v.endswith(accountClass) or v.endswith("AccountsAuthenticationPolicy"):
                policy = m.group(1)

    if settings["storage_backend"] == "kinto.core.storage.memory":  # pragma: no cover
        error_msg = (
            "\033[1;31;40m"
            "The account plugin works really poorly with the memory backend because "
            "accounts are flushed at each startup.\nThis is why you can't use the "
            "kinto create-user command with it."
            "\033[0;37;40m"
        )
        print(error_msg, file=sys.stderr)

    if not policy:
        error_msg = (
            "Account policy missing the 'multiauth.policy.*.use' "
            f"setting. See {accountClass} in docs {DOCS_URL}."
        )
        raise ConfigurationError(error_msg)

    # Add some safety to avoid weird behaviour with basicauth default policy.
    auth_policies = settings["multiauth.policies"]
    if "basicauth" in auth_policies and policy in auth_policies:
        if auth_policies.index("basicauth") < auth_policies.index(policy):
            error_msg = (
                "'basicauth' should not be mentioned before '%s' "
                "in 'multiauth.policies' setting."
            ) % policy
            raise ConfigurationError(error_msg)

    # We assume anyone in account_create_principals is to create
    # accounts for other people.
    # No one can create accounts for other people unless they are an
    # "admin", defined as someone matching account_write_principals.
    # Therefore any account that is in account_create_principals
    # should be in account_write_principals too.
    creators = set(settings.get("account_create_principals", "").split())
    admins = set(settings.get("account_write_principals", "").split())
    cant_create_anything = creators.difference(admins)
    # system.Everyone isn't an account.
    cant_create_anything.discard("system.Everyone")
    if cant_create_anything:
        message = (
            "Configuration has some principals in account_create_principals "
            "but not in account_write_principals. These principals will only be "
            "able to create their own accounts. This may not be what you want.\n"
            "If you want these users to be able to create accounts for other users, "
            "add them to account_write_principals.\n"
            f"Affected users: {list(cant_create_anything)}"
        )

        raise ConfigurationError(message)
